"""This module handles the ontology creation, update and upload to a DSP server. This includes the creation and update
of the project, the creation of groups, users, lists, resource classes, properties and cardinalities."""

from pathlib import Path
from typing import Any
from typing import Optional
from typing import cast

import regex
from loguru import logger

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.commands.excel2json.lists import expand_lists_from_excel
from dsp_tools.commands.project.create.project_create_lists import create_lists_on_server
from dsp_tools.commands.project.create.project_validate import validate_project
from dsp_tools.commands.project.models.context import Context
from dsp_tools.commands.project.models.group import Group
from dsp_tools.commands.project.models.helpers import Cardinality
from dsp_tools.commands.project.models.ontology import Ontology
from dsp_tools.commands.project.models.project import Project
from dsp_tools.commands.project.models.project_definition import ProjectDefinition
from dsp_tools.commands.project.models.propertyclass import PropertyClass
from dsp_tools.commands.project.models.resourceclass import ResourceClass
from dsp_tools.commands.project.models.user import User
from dsp_tools.models.datetimestamp import DateTimeStamp
from dsp_tools.models.exceptions import BaseError
from dsp_tools.models.exceptions import UserError
from dsp_tools.models.langstring import LangString
from dsp_tools.utils.connection import Connection
from dsp_tools.utils.connection_live import ConnectionLive
from dsp_tools.utils.shared import parse_json_input


def _create_project_on_server(
    project_definition: ProjectDefinition,
    con: Connection,
) -> tuple[Project, bool]:
    """
    Create the project on the DSP server.
    If it already exists: update its longname, description, and keywords.

    Args:
        project_definition: object with information about the project
        con: connection to the DSP server

    Raises:
        UserError: if the project cannot be created on the DSP server

    Returns:
        a tuple of the remote project and the success status (True if everything went smoothly, False otherwise)
    """
    all_projects = Project.getAllProjects(con=con)
    if project_definition.shortcode in [proj.shortcode for proj in all_projects]:
        msg = (
            f"The project with the shortcode '{project_definition.shortcode}' already exists on the server.\n"
            f"No changes were made to the project metadata.\n"
            f"Continue with the upload of lists and ontologies ..."
        )
        print(f"WARNING: {msg}")
        logger.warning(msg)

    success = True
    project_local = Project(
        con=con,
        shortcode=project_definition.shortcode,
        shortname=project_definition.shortname,
        longname=project_definition.longname,
        description=LangString(project_definition.descriptions),  # type: ignore[arg-type]
        keywords=set(project_definition.keywords) if project_definition.keywords else None,
        selfjoin=False,
        status=True,
    )
    try:
        project_remote = project_local.create()
    except BaseError:
        err_msg = (
            f"Cannot create project '{project_definition.shortname}' "
            f"({project_definition.shortcode}) on DSP server."
        )
        logger.opt(exception=True).error(err_msg)
        raise UserError(err_msg) from None
    print(f"    Created project '{project_remote.shortname}' ({project_remote.shortcode}).")
    logger.info(f"Created project '{project_remote.shortname}' ({project_remote.shortcode}).")
    return project_remote, success


def _create_groups(
    con: Connection,
    groups: list[dict[str, str]],
    project: Project,
) -> tuple[dict[str, Group], bool]:
    """
    Creates groups on a DSP server from the "groups" section of a JSON project file. If a group cannot be created, it is
    skipped and a warning is printed, but such a group will still be part of the returned dict.
    Returns a tuple consisting of a dict and a bool. The dict contains the groups that have successfully been created
    (or already exist). The bool indicates if everything went smoothly during the process. If a warning or error
    occurred, it is False.

    Args:
        con: connection instance to connect to the DSP server
        groups: "groups" section of a parsed JSON project file
        project: Project the group(s) should be added to (must exist on DSP server)

    Returns:
        A tuple consisting of a dict and the success status.
        The dict has the form ``{group name: group object}``
        for all groups that have successfully been created (or already exist).
        The dict is empty if no group was created.
    """
    overall_success = True
    current_project_groups: dict[str, Group] = {}
    try:
        remote_groups = Group.getAllGroupsForProject(con=con, proj_iri=str(project.iri))
    except BaseError:
        err_msg = (
            "Unable to check if group names are already existing on DSP server, because it is "
            "not possible to retrieve the remote groups from the DSP server."
        )
        print(f"WARNING: {err_msg}")
        logger.opt(exception=True).warning(err_msg)
        remote_groups = []
        overall_success = False

    for group in groups:
        group_name = group["name"]

        # if the group already exists, add it to "current_project_groups" (for later usage), then skip it
        if remotely_existing_group := [g for g in remote_groups if g.name == group_name]:
            current_project_groups[group_name] = remotely_existing_group[0]
            err_msg = f"Group name '{group_name}' already exists on the DSP server. Skipping..."
            print(f"    WARNING: {err_msg}")
            logger.opt(exception=True).warning(err_msg)
            overall_success = False
            continue

        # create the group
        group_local = Group(
            con=con,
            name=group_name,
            descriptions=LangString(group["descriptions"]),
            project=project,
            status=bool(group.get("status", True)),
            selfjoin=bool(group.get("selfjoin", False)),
        )
        try:
            group_remote: Group = group_local.create()
        except BaseError:
            err_msg = "Unable to create group '{group_name}'."
            print(f"    WARNING: {err_msg}")
            logger.opt(exception=True).warning(err_msg)
            overall_success = False
            continue

        current_project_groups[str(group_remote.name)] = group_remote
        print(f"    Created group '{group_name}'.")
        logger.info(f"Created group '{group_name}'.")

    return current_project_groups, overall_success


def _get_group_iris_for_user(
    json_user_definition: dict[str, str],
    current_project: Project,
    current_project_groups: dict[str, Group],
    con: Connection,
    verbose: bool,
) -> tuple[set[str], bool, bool]:
    """
    Retrieve the IRIs of the groups that the user belongs to.

    Args:
        json_user_definition: the section of the JSON file that defines a user
        current_project: the Project object
        current_project_groups: dict of the form ``{group name: group object}``
            with the groups that exist on the DSP server
        con: connection to the DSP server
        verbose: verbose switch

    Returns:
        a tuple consisting of the group IRIs,
        the system admin status (True if the user is sysadmin, False otherwise),
        and the success status (True if everything went well)

    Raises:
        BaseError: if no groups can be retrieved from the DSP server, or if the retrieved group has no IRI
    """
    success = True
    username = json_user_definition["username"]
    group_iris: set[str] = set()
    sysadmin = False
    remote_groups: list[Group] = []
    for full_group_name in json_user_definition.get("groups", []):
        # full_group_name has the form '[project_shortname]:group_name' or 'SystemAdmin'
        inexisting_group_msg = (
            f"User {username} cannot be added to group {full_group_name}, because such a group doesn't exist."
        )
        if ":" not in full_group_name and full_group_name != "SystemAdmin":
            print(f"    WARNING: {inexisting_group_msg}")
            logger.opt(exception=True).warning(inexisting_group_msg)
            success = False
            continue

        if full_group_name == "SystemAdmin":
            sysadmin = True
            if verbose:
                print(f"    Added user '{username}' to group 'SystemAdmin'.")
            logger.info(f"Added user '{username}' to group 'SystemAdmin'.")
            continue

        # all other cases (":" in full_group_name)
        project_shortname, group_name = full_group_name.split(":")
        if not project_shortname:
            # full_group_name refers to a group inside the same project
            if group_name not in current_project_groups:
                print(f"    WARNING: {inexisting_group_msg}")
                logger.opt(exception=True).warning(inexisting_group_msg)
                success = False
                continue
            group = current_project_groups[group_name]
        else:
            # full_group_name refers to an already existing group on DSP
            try:
                # "remote_groups" might be available from a previous loop cycle
                remote_groups = remote_groups or Group.getAllGroups(con=con)
            except BaseError:
                err_msg = (
                    f"User '{username}' is referring to the group {full_group_name} that "
                    f"exists on the DSP server, but no groups could be retrieved from the DSP server."
                )
                print(f"    WARNING: {err_msg}")
                logger.opt(exception=True).warning(err_msg)
                success = False
                continue
            existing_group = [g for g in remote_groups if g.project == current_project.iri and g.name == group_name]
            if not existing_group:
                print(f"    WARNING: {inexisting_group_msg}")
                logger.opt(exception=True).warning(inexisting_group_msg)
                success = False
                continue
            group = existing_group[0]

        if not group.iri:
            raise BaseError(f"Group '{group}' has no IRI.")
        group_iris.add(group.iri)
        if verbose:
            print(f"    Added user '{username}' to group '{full_group_name}'.")
        logger.info(f"Added user '{username}' to group '{full_group_name}'.")

    return group_iris, sysadmin, success


def _get_projects_where_user_is_admin(
    json_user_definition: dict[str, str],
    current_project: Project,
    con: Connection,
    verbose: bool,
) -> tuple[dict[str, bool], bool]:
    """
    Create a dict that tells for every project if the user is administrator in that project or not.

    Args:
        json_user_definition: the section of the JSON file that defines a user
        current_project: the Project object
        con: connection to the DSP server
        verbose: verbose switch

    Returns:
        a tuple consisting of a dict in the form {project IRI: isAdmin}, and the success status
    """
    success = True
    username = json_user_definition["username"]
    project_info: dict[str, bool] = {}
    remote_projects: list[Project] = []
    for full_project_name in json_user_definition.get("projects", []):
        # full_project_name has the form '[project_name]:member' or '[project_name]:admin'
        if ":" not in full_project_name:
            err_msg = "Provided project '{full_project_name}' for user '{username}' is not valid. Skipping..."
            print(f"    WARNING: {err_msg}")
            logger.opt(exception=True).warning(err_msg)
            success = False
            continue

        project_name, project_role = full_project_name.split(":")
        if not project_name:
            # full_project_name refers to the current project
            in_project = current_project
        else:
            # full_project_name refers to an already existing project on DSP
            try:
                # "remote_projects" might be available from a previous loop cycle
                remote_projects = remote_projects or current_project.getAllProjects(con=con)
            except BaseError:
                err_msg = (
                    f"User '{username}' cannot be added to the projects {json_user_definition['projects']} "
                    f"because the projects cannot be retrieved from the DSP server."
                )
                print(f"    WARNING: {err_msg}")
                logger.opt(exception=True).warning(err_msg)
                success = False
                continue
            in_project_list = [p for p in remote_projects if p.shortname == project_name]
            if not in_project_list:
                msg = f"Provided project '{full_project_name}' for user '{username}' is not valid. Skipping..."
                print(f"    WARNING: {msg}")
                logger.opt(exception=True).warning(msg)
                success = False
                continue
            in_project = in_project_list[0]

        is_admin = project_role == "admin"
        project_info[str(in_project.iri)] = is_admin
        if verbose:
            print(f"    Added user '{username}' as {project_role} to project '{in_project.shortname}'.")
        logger.info(f"Added user '{username}' as {project_role} to project '{in_project.shortname}'.")

    return project_info, success


def _create_users(
    con: Connection,
    users_section: list[dict[str, str]],
    current_project_groups: dict[str, Group],
    current_project: Project,
    verbose: bool,
) -> bool:
    """
    Creates users on a DSP server from the "users" section of a JSON project file.
    If a user cannot be created, a warning is printed and the user is skipped.

    Args:
        con: connection instance to connect to the DSP server
        users_section: "users" section of a parsed JSON project file
        current_project_groups: groups defined in the current project, in the form ``{group name: group object}``
            (must exist on DSP server)
        current_project: "project" object of the current project (must exist on DSP server)
        verbose: Prints more information if set to True

    Returns:
        True if all users could be created without any problems. False if a warning/error occurred.
    """
    overall_success = True
    for json_user_definition in users_section:
        username = json_user_definition["username"]

        # skip the user if he already exists
        all_users = User.getAllUsers(con)
        if json_user_definition["email"] in [user.email for user in all_users]:
            err_msg = (
                f"User '{username}' already exists on the DSP server.\n"
                f"Please manually add this user to the project in DSP-APP."
            )
            print(f"    WARNING: {err_msg}")
            logger.opt(exception=True).warning(err_msg)
            overall_success = False
            continue
        # add user to the group(s)
        group_iris, sysadmin, success = _get_group_iris_for_user(
            json_user_definition=json_user_definition,
            current_project=current_project,
            current_project_groups=current_project_groups,
            con=con,
            verbose=verbose,
        )
        if not success:
            overall_success = False

        # add user to the project(s)
        project_info, success = _get_projects_where_user_is_admin(
            json_user_definition=json_user_definition,
            current_project=current_project,
            con=con,
            verbose=verbose,
        )
        if not success:
            overall_success = False

        # create the user
        user_local = User(
            con=con,
            username=json_user_definition["username"],
            email=json_user_definition["email"],
            givenName=json_user_definition["givenName"],
            familyName=json_user_definition["familyName"],
            password=json_user_definition["password"],
            status=bool(json_user_definition.get("status", True)),
            lang=json_user_definition.get("lang", "en"),
            sysadmin=sysadmin,
            in_projects=project_info,
            in_groups=group_iris,
        )
        try:
            user_local.create()
        except BaseError:
            print(f"    WARNING: Unable to create user '{username}'.")
            logger.opt(exception=True).warning(f"Unable to create user '{username}'.")
            overall_success = False
            continue
        print(f"    Created user '{username}'.")
        logger.info(f"Created user '{username}'.")

    return overall_success


def _sort_resources(
    unsorted_resources: list[dict[str, Any]],
    onto_name: str,
) -> list[dict[str, Any]]:
    """
    This method sorts the resource classes in an ontology according to their inheritance order (parent classes first).

    Args:
        unsorted_resources: list of resources from a parsed JSON project file
        onto_name: name of the onto

    Returns:
        sorted list of resource classes
    """

    # do not modify the original unsorted_resources, which points to the original JSON project file
    resources_to_sort = unsorted_resources.copy()
    sorted_resources: list[dict[str, Any]] = []
    ok_resource_names: list[str] = []
    while resources_to_sort:
        # inside the for loop, resources_to_sort is modified, so a copy must be made to iterate over
        for res in resources_to_sort.copy():
            parent_classes = res["super"]
            if isinstance(parent_classes, str):
                parent_classes = [parent_classes]
            parent_classes = [regex.sub(r"^:([^:]+)$", f"{onto_name}:\\1", elem) for elem in parent_classes]
            parent_classes_ok = [not p.startswith(onto_name) or p in ok_resource_names for p in parent_classes]
            if all(parent_classes_ok):
                sorted_resources.append(res)
                res_name = f'{onto_name}:{res["name"]}'
                ok_resource_names.append(res_name)
                resources_to_sort.remove(res)
    return sorted_resources


def _sort_prop_classes(
    unsorted_prop_classes: list[dict[str, Any]],
    onto_name: str,
) -> list[dict[str, Any]]:
    """
    In case of inheritance, parent properties must be uploaded before their children. This method sorts the
    properties.

    Args:
        unsorted_prop_classes: list of properties from a parsed JSON project file
        onto_name: name of the onto

    Returns:
        sorted list of properties
    """

    # do not modify the original unsorted_prop_classes, which points to the original JSON project file
    prop_classes_to_sort = unsorted_prop_classes.copy()
    sorted_prop_classes: list[dict[str, Any]] = []
    ok_propclass_names: list[str] = []
    while prop_classes_to_sort:
        # inside the for loop, resources_to_sort is modified, so a copy must be made to iterate over
        for prop in prop_classes_to_sort.copy():
            prop_name = f'{onto_name}:{prop["name"]}'
            parent_classes = prop.get("super", "hasValue")
            if isinstance(parent_classes, str):
                parent_classes = [parent_classes]
            parent_classes = [regex.sub(r"^:([^:]+)$", f"{onto_name}:\\1", elem) for elem in parent_classes]
            parent_classes_ok = [not p.startswith(onto_name) or p in ok_propclass_names for p in parent_classes]
            if all(parent_classes_ok):
                sorted_prop_classes.append(prop)
                ok_propclass_names.append(prop_name)
                prop_classes_to_sort.remove(prop)
    return sorted_prop_classes


def _create_ontology(
    onto_name: str,
    onto_label: str,
    onto_comment: Optional[str],
    project_ontologies: list[Ontology],
    con: Connection,
    project_remote: Project,
    context: Context,
    verbose: bool,
) -> Optional[Ontology]:
    """
    Create an ontology on the DSP server,
    and add the prefixes defined in the JSON file to its context.
    If the ontology already exists on the DSP server, it is skipped.

    Args:
        onto_name: name of the ontology
        onto_label: label of the ontology
        onto_comment: comment of the ontology
        project_ontologies: ontologies existing on the DSP server
        con: Connection to the DSP server
        project_remote: representation of the project on the DSP server
        context: prefixes and the ontology IRIs they stand for
        verbose: verbose switch

    Raises:
        UserError: if the ontology cannot be created on the DSP server

    Returns:
        representation of the created ontology on the DSP server, or None if it already existed
    """
    # skip if it already exists on the DSP server
    if onto_name in [onto.name for onto in project_ontologies]:
        err_msg = f"Ontology '{onto_name}' already exists on the DSP server. Skipping..."
        print(f"    WARNING: {err_msg}")
        logger.opt(exception=True).warning(err_msg)
        return None

    print(f"Create ontology '{onto_name}'...")
    logger.info(f"Create ontology '{onto_name}'...")
    ontology_local = Ontology(
        con=con,
        project=project_remote,
        label=onto_label,
        name=onto_name,
        comment=onto_comment,
    )
    try:
        ontology_remote = ontology_local.create()
    except BaseError:
        # if ontology cannot be created, let the error escalate
        logger.opt(exception=True).error(f"ERROR while trying to create ontology '{onto_name}'.")
        raise UserError(f"ERROR while trying to create ontology '{onto_name}'.") from None

    if verbose:
        print(f"    Created ontology '{onto_name}'.")
    logger.info(f"Created ontology '{onto_name}'.")

    context.add_context(
        ontology_remote.name,
        ontology_remote.iri + ("" if ontology_remote.iri.endswith("#") else "#"),
    )

    # add the prefixes defined in the JSON file
    for onto_prefix, onto_info in context:
        if onto_info and str(onto_prefix) not in ontology_remote.context:
            onto_iri = onto_info.iri + ("#" if onto_info.hashtag else "")
            ontology_remote.context.add_context(prefix=str(onto_prefix), iri=onto_iri)

    return ontology_remote


def _create_ontologies(
    con: Connection,
    context: Context,
    knora_api_prefix: str,
    names_and_iris_of_list_nodes: dict[str, Any],
    ontology_definitions: list[dict[str, Any]],
    project_remote: Project,
    verbose: bool,
) -> bool:
    """
    Iterates over the ontologies in a JSON project file and creates the ontologies that don't exist on the DSP server
    yet. For every ontology, it first creates the resource classes, then the properties, and then adds the cardinalities
    to the resource classes.

    Args:
        con: Connection to the DSP server
        context: prefixes and the ontology IRIs they stand for
        knora_api_prefix: the prefix that stands for the knora-api ontology
        names_and_iris_of_list_nodes: IRIs of list nodes that were already created and are available on the DSP server
        ontology_definitions: the "ontologies" section of the parsed JSON project file
        project_remote: representation of the project on the DSP server
        verbose: verbose switch

    Raises:
        UserError: if an error occurs during the creation of an ontology.
        All other errors are printed, the process continues, but the success status will be false.

    Returns:
        True if everything went smoothly, False otherwise
    """

    overall_success = True

    print("Create ontologies...")
    logger.info("Create ontologies...")
    try:
        project_ontologies = Ontology.getProjectOntologies(con=con, project_id=str(project_remote.iri))
    except BaseError:
        err_msg = "Unable to retrieve remote ontologies. Cannot check if your ontology already exists."
        print("WARNING: {err_msg}")
        logger.opt(exception=True).warning(err_msg)
        project_ontologies = []

    for ontology_definition in ontology_definitions:
        ontology_remote = _create_ontology(
            onto_name=ontology_definition["name"],
            onto_label=ontology_definition["label"],
            onto_comment=ontology_definition.get("comment"),
            project_ontologies=project_ontologies,
            con=con,
            project_remote=project_remote,
            context=context,
            verbose=verbose,
        )
        if not ontology_remote:
            overall_success = False
            continue

        # add the empty resource classes to the remote ontology
        last_modification_date, remote_res_classes, success = _add_resource_classes_to_remote_ontology(
            onto_name=ontology_definition["name"],
            resclass_definitions=ontology_definition.get("resources", []),
            ontology_remote=ontology_remote,
            con=con,
            last_modification_date=ontology_remote.lastModificationDate,
            verbose=verbose,
        )
        if not success:
            overall_success = False

        # add the property classes to the remote ontology
        last_modification_date, success = _add_property_classes_to_remote_ontology(
            onto_name=ontology_definition["name"],
            property_definitions=ontology_definition.get("properties", []),
            ontology_remote=ontology_remote,
            names_and_iris_of_list_nodes=names_and_iris_of_list_nodes,
            con=con,
            last_modification_date=last_modification_date,
            knora_api_prefix=knora_api_prefix,
            verbose=verbose,
        )
        if not success:
            overall_success = False

        # Add cardinalities to class
        success = _add_cardinalities_to_resource_classes(
            resclass_definitions=ontology_definition.get("resources", []),
            ontology_remote=ontology_remote,
            remote_res_classes=remote_res_classes,
            last_modification_date=last_modification_date,
            knora_api_prefix=knora_api_prefix,
            verbose=verbose,
        )
        if not success:
            overall_success = False

    return overall_success


def _add_resource_classes_to_remote_ontology(
    onto_name: str,
    resclass_definitions: list[dict[str, Any]],
    ontology_remote: Ontology,
    con: Connection,
    last_modification_date: DateTimeStamp,
    verbose: bool,
) -> tuple[DateTimeStamp, dict[str, ResourceClass], bool]:
    """
    Creates the resource classes (without cardinalities) defined in the "resources" section of an ontology. The
    containing project and the containing ontology must already be existing on the DSP server.
    If an error occurs during creation of a resource class, it is printed out, the process continues, but the success
    status will be false.

    Args:
        onto_name: name of the current ontology
        resclass_definitions: the part of the parsed JSON project file that contains the resources of the current onto
        ontology_remote: representation of the current ontology on the DSP server
        con: connection to the DSP server
        last_modification_date: last modification date of the ontology on the DSP server
        verbose: verbose switch

    Returns:
        last modification date of the ontology,
        new resource classes,
        success status
    """

    overall_success = True
    print("    Create resource classes...")
    logger.info("Create resource classes...")
    new_res_classes: dict[str, ResourceClass] = {}
    sorted_resources = _sort_resources(resclass_definitions, onto_name)
    for res_class in sorted_resources:
        super_classes = res_class["super"]
        if isinstance(super_classes, str):
            super_classes = [super_classes]
        res_class_local = ResourceClass(
            con=con,
            context=ontology_remote.context,
            ontology_id=ontology_remote.iri,
            name=res_class["name"],
            superclasses=super_classes,
            label=LangString(res_class.get("labels")),
            comment=LangString(res_class.get("comments")) if res_class.get("comments") else None,
        )
        try:
            last_modification_date, res_class_remote = res_class_local.create(last_modification_date)
            new_res_classes[str(res_class_remote.iri)] = res_class_remote
            ontology_remote.lastModificationDate = last_modification_date
            if verbose:
                print(f"    Created resource class '{res_class['name']}'")
            logger.info(f"Created resource class '{res_class['name']}'")
        except BaseError:
            err_msg = f"Unable to create resource class '{res_class['name']}'."
            print(f"WARNING: {err_msg}")
            logger.opt(exception=True).warning(err_msg)
            overall_success = False

    return last_modification_date, new_res_classes, overall_success


def _add_property_classes_to_remote_ontology(
    onto_name: str,
    property_definitions: list[dict[str, Any]],
    ontology_remote: Ontology,
    names_and_iris_of_list_nodes: dict[str, Any],
    con: Connection,
    last_modification_date: DateTimeStamp,
    knora_api_prefix: str,
    verbose: bool,
) -> tuple[DateTimeStamp, bool]:
    """
    Creates the property classes defined in the "properties" section of an ontology. The
    containing project and the containing ontology must already be existing on the DSP server.
    If an error occurs during creation of a property class, it is printed out, the process continues, but the success
    status will be false.

    Args:
        onto_name: name of the current ontology
        property_definitions: the part of the parsed JSON project file that contains the properties of the current onto
        ontology_remote: representation of the current ontology on the DSP server
        names_and_iris_of_list_nodes: IRIs of list nodes that were already created and are available on the DSP server
        con: connection to the DSP server
        last_modification_date: last modification date of the ontology on the DSP server
        knora_api_prefix: the prefix that stands for the knora-api ontology
        verbose: verbose switch

    Returns:
        a tuple consisting of the last modification date of the ontology, and the success status
    """
    overall_success = True
    print("    Create property classes...")
    logger.info("Create property classes...")
    sorted_prop_classes = _sort_prop_classes(property_definitions, onto_name)
    for prop_class in sorted_prop_classes:
        # get the super-property/ies, valid forms are:
        #   - "prefix:super-property" : fully qualified name of property in another ontology. The prefix has to be
        #     defined in the prefixes part.
        #   - ":super-property" : super-property defined in current ontology
        #   - "super-property" : super-property defined in the knora-api ontology
        #   - if omitted, "knora-api:hasValue" is assumed
        if prop_class.get("super"):
            super_props = []
            for super_class in prop_class["super"]:
                if ":" in super_class:
                    prefix, _class = super_class.split(":")
                    super_props.append(super_class if prefix else f"{ontology_remote.name}:{_class}")
                else:
                    super_props.append(knora_api_prefix + super_class)
        else:
            super_props = ["knora-api:hasValue"]

        # get the "object", valid forms are:
        #   - "prefix:object_name" : fully qualified object. The prefix has to be defined in the prefixes part.
        #   - ":object_name" : The object is defined in the current ontology.
        #   - "object_name" : The object is defined in "knora-api"
        if ":" in prop_class["object"]:
            prefix, _object = prop_class["object"].split(":")
            prop_object = f"{prefix}:{_object}" if prefix else f"{ontology_remote.name}:{_object}"
        else:
            prop_object = knora_api_prefix + prop_class["object"]

        # get the gui_attributes
        gui_attributes = prop_class.get("gui_attributes")
        if gui_attributes and gui_attributes.get("hlist"):
            list_iri = names_and_iris_of_list_nodes[gui_attributes["hlist"]]["id"]
            gui_attributes["hlist"] = f"<{list_iri}>"

        # create the property class
        prop_class_local = PropertyClass(
            con=con,
            context=ontology_remote.context,
            label=LangString(prop_class.get("labels")),
            name=prop_class["name"],
            ontology_id=ontology_remote.iri,
            superproperties=super_props,
            rdf_object=prop_object,
            rdf_subject=prop_class.get("subject"),
            gui_element="salsah-gui:" + prop_class["gui_element"],
            gui_attributes=gui_attributes,
            comment=LangString(prop_class["comments"]) if prop_class.get("comments") else None,
        )
        try:
            last_modification_date, _ = prop_class_local.create(last_modification_date)
            ontology_remote.lastModificationDate = last_modification_date
            if verbose:
                print(f"    Created property class '{prop_class['name']}'")
            logger.info(f"Created property class '{prop_class['name']}'")
        except BaseError:
            err_msg = f"Unable to create property class '{prop_class['name']}'."
            print(f"WARNING: {err_msg}")
            logger.opt(exception=True).warning(f"Unable to create property class '{prop_class['name']}'.")
            overall_success = False

    return last_modification_date, overall_success


def _add_cardinalities_to_resource_classes(
    resclass_definitions: list[dict[str, Any]],
    ontology_remote: Ontology,
    remote_res_classes: dict[str, ResourceClass],
    last_modification_date: DateTimeStamp,
    knora_api_prefix: str,
    verbose: bool,
) -> bool:
    """
    Iterates over the resource classes of an ontology of a JSON project definition, and adds the cardinalities to each
    resource class. The resource classes and the properties must already be existing on the DSP server.
    If an error occurs during creation of a cardinality, it is printed out, the process continues, but the success
    status will be false.

    Args:
        resclass_definitions: the part of the parsed JSON project file that contains the resources of the current onto
        ontology_remote: representation of the current ontology on the DSP server
        remote_res_classes: representations of the resource classes on the DSP server
        last_modification_date: last modification date of the ontology on the DSP server
        knora_api_prefix: the prefix that stands for the knora-api ontology
        verbose: verbose switch

    Returns:
        success status
    """
    overall_success = True
    print("    Add cardinalities to resource classes...")
    logger.info("Add cardinalities to resource classes...")
    switcher = {
        "1": Cardinality.C_1,
        "0-1": Cardinality.C_0_1,
        "0-n": Cardinality.C_0_n,
        "1-n": Cardinality.C_1_n,
    }
    for res_class in resclass_definitions:
        res_class_remote = remote_res_classes.get(f"{ontology_remote.iri}#{res_class['name']}")
        if not res_class_remote:
            msg = (
                f"Unable to add cardinalities to resource class '{res_class['name']}': "
                f"This class doesn't exist on the DSP server."
            )
            print(f"WARNINIG: {msg}")
            logger.opt(exception=True).warning(msg)
            overall_success = False
            continue
        for card_info in res_class.get("cardinalities", []):
            if ":" in card_info["propname"]:
                prefix, prop = card_info["propname"].split(":")
                qualified_propname = card_info["propname"] if prefix else f"{ontology_remote.name}:{prop}"
            else:
                qualified_propname = knora_api_prefix + card_info["propname"]

            try:
                last_modification_date = res_class_remote.addProperty(
                    property_id=qualified_propname,
                    cardinality=switcher[card_info["cardinality"]],
                    gui_order=card_info.get("gui_order"),
                    last_modification_date=last_modification_date,
                )
                if verbose:
                    print(f"    Added cardinality '{card_info['propname']}' to resource class '{res_class['name']}'")
                logger.info(f"Added cardinality '{card_info['propname']}' to resource class '{res_class['name']}'")
            except BaseError:
                err_msg = f"Unable to add cardinality '{qualified_propname}' to resource class {res_class['name']}."
                print(f"WARNING: {err_msg}")
                logger.opt(exception=True).warning(err_msg)
                overall_success = False

            ontology_remote.lastModificationDate = last_modification_date

    return overall_success


def _rectify_hlist_of_properties(
    lists: list[dict[str, Any]],
    properties: list[dict[str, Any]],
) -> list[dict[str, Any]]:
    """
    Check the "hlist" of the "gui_attributes" of the properties.
    If they don't refer to an existing list name,
    check if there is a label of a list that corresponds to the "hlist".
    If so, rectify the "hlist" to refer to the name of the list instead of the label.

    Args:
        lists: "lists" section of the JSON project definition
        properties: "properties" section of one of the ontologies of the JSON project definition

    Raises:
        UserError: if the "hlist" refers to no existing list name or label

    Returns:
        the rectified "properties" section
    """

    if not lists or not properties:
        return properties

    existing_list_names = [lst["name"] for lst in lists]

    for prop in properties:
        if not prop.get("gui_attributes"):
            continue
        if not prop["gui_attributes"].get("hlist"):
            continue
        list_name = prop["gui_attributes"]["hlist"] if prop["gui_attributes"]["hlist"] in existing_list_names else None
        if list_name:
            continue

        deduced_list_name = None
        for root_node in lists:
            if prop["gui_attributes"]["hlist"] in root_node["labels"].values():
                deduced_list_name = cast(str, root_node["name"])
        if deduced_list_name:
            msg = (
                f"INFO: Property '{prop['name']}' references the list '{prop['gui_attributes']['hlist']}' "
                f"which is not a valid list name. "
                f"Assuming that you meant '{deduced_list_name}' instead."
            )
            logger.opt(exception=True).warning(msg)
            print(msg)
        else:
            msg = f"Property '{prop['name']}' references an unknown list: '{prop['gui_attributes']['hlist']}'"
            logger.error(msg)
            raise UserError(f"ERROR: {msg}")
        prop["gui_attributes"]["hlist"] = deduced_list_name

    return properties


def create_project(
    project_file_as_path_or_parsed: str | Path | dict[str, Any],
    creds: ServerCredentials,
    verbose: bool = False,
) -> bool:
    """
    Creates a project from a JSON project file on a DSP server.
    A project must contain at least one ontology,
    and it may contain lists, users, and groups.
    Severe errors lead to a BaseError,
    while other errors are printed without interrupting the process.

    Args:
        project_file_as_path_or_parsed: path to the JSON project definition, or parsed JSON object
        creds: credentials to connect to the DSP server
        verbose: prints more information if set to True

    Raises:
        UserError:
          - if the project cannot be created
          - if the login fails
          - if an ontology cannot be created

        BaseError:
          - if the input is invalid
          - if an Excel file referenced in the "lists" section cannot be expanded
          - if the validation doesn't pass

    Returns:
        True if everything went smoothly, False if a warning or error occurred
    """

    knora_api_prefix = "knora-api:"
    overall_success = True

    project_json = parse_json_input(project_file_as_path_or_parsed=project_file_as_path_or_parsed)

    context = Context(project_json.get("prefixes", {}))

    project_definition = _prepare_and_validate_project(project_json)

    all_lists = _get_all_lists(project_json)

    all_ontos = _get_all_ontos(project_json, all_lists)

    # establish connection to DSP server
    con = ConnectionLive(creds.server)
    con.login(creds.user, creds.password)

    # create project on DSP server
    info_str = f"Create project '{project_definition.shortname}' ({project_definition.shortcode})..."
    print(info_str)
    logger.info(info_str)
    project_remote, success = _create_project_on_server(
        project_definition=project_definition,
        con=con,
    )
    if not success:
        overall_success = False

    # create the lists
    names_and_iris_of_list_nodes: dict[str, Any] = {}
    if all_lists:
        print("Create lists...")
        logger.info("Create lists...")
        names_and_iris_of_list_nodes, success = create_lists_on_server(
            lists_to_create=all_lists,
            con=con,
            project_remote=project_remote,
        )
        if not success:
            overall_success = False

    # create the groups
    current_project_groups: dict[str, Group] = {}
    if project_definition.groups:
        print("Create groups...")
        logger.info("Create groups...")
        current_project_groups, success = _create_groups(
            con=con,
            groups=project_definition.groups,
            project=project_remote,
        )
        if not success:
            overall_success = False

    # create or update the users
    if project_definition.users:
        print("Create users...")
        logger.info("Create users...")
        success = _create_users(
            con=con,
            users_section=project_definition.users,
            current_project_groups=current_project_groups,
            current_project=project_remote,
            verbose=verbose,
        )
        if not success:
            overall_success = False

    # create the ontologies
    success = _create_ontologies(
        con=con,
        context=context,
        knora_api_prefix=knora_api_prefix,
        names_and_iris_of_list_nodes=names_and_iris_of_list_nodes,
        ontology_definitions=all_ontos,
        project_remote=project_remote,
        verbose=verbose,
    )
    if not success:
        overall_success = False

    # final steps
    if overall_success:
        msg = (
            f"Successfully created project '{project_definition.shortname}' "
            f"({project_definition.shortcode}) with all its ontologies. "
            f"There were no problems during the creation process."
        )
        print(f"========================================================\n{msg}")
        logger.info(msg)
    else:
        msg = (
            f"The project '{project_definition.shortname}' ({project_definition.shortcode}) "
            f"with its ontologies could be created, "
            f"but during the creation process, some problems occurred. Please carefully check the console output."
        )
        print(f"========================================================\nWARNING: {msg}")
        logger.opt(exception=True).warning(msg)

    return overall_success


def _prepare_and_validate_project(
    project_json: dict[str, Any],
) -> ProjectDefinition:
    project_def = ProjectDefinition(
        shortcode=project_json["project"]["shortcode"],
        shortname=project_json["project"]["shortname"],
        longname=project_json["project"]["longname"],
        keywords=project_json["project"].get("keywords"),
        descriptions=project_json["project"].get("descriptions"),
        groups=project_json["project"].get("groups"),
        users=project_json["project"].get("users"),
    )

    # validate against JSON schema
    validate_project(project_json, expand_lists=False)
    print("    JSON project file is syntactically correct and passed validation.")
    logger.info("JSON project file is syntactically correct and passed validation.")

    return project_def


def _get_all_lists(project_json: dict[str, Any]) -> list[dict[str, Any]] | None:
    # expand the Excel files referenced in the "lists" section of the project, if any
    if all_lists := expand_lists_from_excel(project_json.get("project", {}).get("lists", [])):
        return all_lists
    new_lists: list[dict[str, Any]] | None = project_json["project"].get("lists")
    return new_lists


def _get_all_ontos(project_json: dict[str, Any], all_lists: list[dict[str, Any]] | None) -> list[dict[str, Any]]:
    all_ontos: list[dict[str, Any]] = project_json["project"]["ontologies"]
    if all_lists is None:
        return all_ontos
    # rectify the "hlist" of the "gui_attributes" of the properties
    for onto in all_ontos:
        if onto.get("properties"):
            onto["properties"] = _rectify_hlist_of_properties(
                lists=all_lists,
                properties=onto["properties"],
            )
    return all_ontos



from dataclasses import dataclass
from typing import Any

import pytest
from pytest_unordered import unordered

from dsp_tools.commands.xmlupload.ontology_client import OntologyClientLive
from test.integration.commands.xmlupload.connection_mock import ConnectionMockBase


@dataclass
class ConnectionMock(ConnectionMockBase):
    get_response: dict[Any, Any]

    def get(self, route: str, headers: dict[str, str] | None = None) -> dict[Any, Any]:  # noqa: ARG002 (unused-method-argument)
        return self.get_response


def test_get_ontology_names_from_server() -> None:
    response = {
        "project": {
            "description": [],
            "id": "http://rdfh.ch/projects/yTerZGyxjZVqFMNNKXCDPF",
            "keywords": [],
            "logo": None,
            "longname": "Bernoulli-Euler Online",
            "ontologies": [
                "http://api.dasch.swiss/ontology/0801/biblio/v2",
                "http://api.dasch.swiss/ontology/0801/newton/v2",
                "http://api.dasch.swiss/ontology/0801/leibniz/v2",
                "http://api.dasch.swiss/ontology/0801/beol/v2",
            ],
            "selfjoin": False,
            "shortcode": "0801",
            "shortname": "beol",
            "status": True,
        }
    }
    con = ConnectionMock(response)
    onto_client = OntologyClientLive(con, "0801", "beol")
    onto_client._get_ontology_names_from_server()
    assert unordered(onto_client.ontology_names) == ["biblio", "newton", "leibniz", "beol"]


def test_get_ontology_from_server() -> None:
    response = {
        "knora-api:lastModificationDate": {},
        "rdfs:label": "The BEOL ontology",
        "@graph": [
            {"resource_class": ["Information"]},
            {"property": ["Information"]},
        ],
        "knora-api:attachedToProject": {},
        "@type": "owl:Ontology",
        "@id": "http://api.dasch.swiss/ontology/0801/beol/v2",
        "@context": {},
    }
    con = ConnectionMock(response)
    onto_client = OntologyClientLive(con, "0801", "beol")
    res_graph = onto_client._get_ontology_from_server("beol")
    assert unordered(res_graph) == [{"resource_class": ["Information"]}, {"property": ["Information"]}]


def test_get_knora_api_from_server() -> None:
    response = {
        "rdfs:label": "The knora-api ontology in the complex schema",
        "@graph": [
            {"resource_class": ["Information"]},
            {"property": ["Information"]},
        ],
        "knora-api:attachedToProject": {},
        "knora-api:isBuiltIn": True,
        "@type": "owl:Ontology",
        "@id": "http://api.knora.org/ontology/knora-api/v2",
        "@context": {},
    }
    con = ConnectionMock(response)
    onto_client = OntologyClientLive(con, "", "")
    res_graph = onto_client.get_knora_api_ontology_from_server()
    assert unordered(res_graph) == [{"resource_class": ["Information"]}, {"property": ["Information"]}]


if __name__ == "__main__":
    pytest.main([__file__])



import warnings
from copy import deepcopy
from pathlib import Path

from lxml import etree
from rdflib import SH
from rdflib import Graph
from termcolor import cprint

from dsp_tools.commands.xml_validate.api_connection import OntologyConnection
from dsp_tools.commands.xml_validate.api_connection import ShaclValidator
from dsp_tools.commands.xml_validate.deserialise_input import deserialise_xml
from dsp_tools.commands.xml_validate.make_data_rdf import make_data_rdf
from dsp_tools.commands.xml_validate.models.data_deserialised import ProjectDeserialised
from dsp_tools.commands.xml_validate.models.data_rdf import DataRDF
from dsp_tools.commands.xml_validate.models.validation import ValidationReport
from dsp_tools.commands.xml_validate.reformat_validaton_result import reformat_validation_graph
from dsp_tools.commands.xml_validate.sparql.construct_shapes import construct_shapes_graph
from dsp_tools.models.custom_warnings import DspToolsUserWarning
from dsp_tools.utils.xml_utils import parse_xml_file
from dsp_tools.utils.xml_utils import remove_comments_from_element_tree
from dsp_tools.utils.xml_utils import transform_into_localnames
from dsp_tools.utils.xml_validation import validate_xml

LIST_SEPARATOR = "\n    - "


def xml_validate(filepath: Path, api_url: str, dev_route: bool) -> bool:  # noqa: ARG001 (unused argument)
    """
    Takes a file and project information and validates it against the ontologies on the server.

    Args:
        filepath: path to the xml data file
        api_url: url of the api host
        dev_route: if this flag is set features that are still in development will be used

    Returns:
        true unless it crashed
    """
    _inform_about_experimental_feature()
    data_rdf, shortcode = _get_data_info_from_file(filepath, api_url)
    onto_con = OntologyConnection(api_url, shortcode)
    ontologies, shapes = _get_shacl(onto_con)
    data_graph = data_rdf.make_graph()
    # data_graph += ontologies
    val = ShaclValidator(api_url)
    report = _validate(val, shapes, data_graph)
    if report.conforms:
        cprint("\n   Validation passed!   ", color="green", attrs=["bold", "reverse"])
    else:
        reformatted = reformat_validation_graph(report.validation_graph, data_graph)
        problem_msg = reformatted.get_msg(filepath)
        cprint("\n   Validation errors found!   ", color="light_red", attrs=["bold", "reverse"])
        print(problem_msg)
        if reformatted.unexpected_results:
            reformatted.unexpected_results.save_inform_user(
                results_graph=report.validation_graph,
                shacl=report.shacl_graph,
                data=data_graph,
            )
    return True


def _inform_about_experimental_feature() -> None:
    what_is_validated = [
        "This is an experimental feature, it will change and be extended continuously. "
        "The following information of your data is being validated:",
        "Cardinalities",
    ]
    warnings.warn(DspToolsUserWarning(LIST_SEPARATOR.join(what_is_validated)))


def _validate(validator: ShaclValidator, shapes_graph: Graph, data_graph: Graph) -> ValidationReport:
    shape_str = shapes_graph.serialize(format="ttl")
    data_str = data_graph.serialize(format="ttl")
    results = validator.validate(data_str, shape_str)
    conforms = bool(next(results.objects(None, SH.conforms)))
    return ValidationReport(
        conforms=conforms,
        validation_graph=results,
        shacl_graph=shapes_graph,
        data_graph=data_graph,
    )


def _get_shacl(onto_con: OntologyConnection) -> tuple[Graph, Graph]:
    ontologies = _get_project_ontos(onto_con)
    knora_ttl = onto_con.get_knora_api()
    kag = Graph()
    kag.parse(data=knora_ttl, format="ttl")
    onto_for_construction = deepcopy(ontologies) + kag
    shapes = construct_shapes_graph(onto_for_construction)
    shapes += ontologies
    return ontologies, shapes


def _get_project_ontos(onto_con: OntologyConnection) -> Graph:
    all_ontos = onto_con.get_ontologies()
    onto_g = Graph()
    for onto in all_ontos:
        og = Graph()
        og.parse(data=onto, format="ttl")
        onto_g += og
    return onto_g


def _get_data_info_from_file(file: Path, api_url: str) -> tuple[DataRDF, str]:
    cleaned_root = _parse_and_clean_file(file, api_url)
    deserialised: ProjectDeserialised = deserialise_xml(cleaned_root)
    rdf_data: DataRDF = make_data_rdf(deserialised.data)
    return rdf_data, deserialised.info.shortcode


def _parse_and_clean_file(file: Path, api_url: str) -> etree._Element:
    root = parse_xml_file(file)
    root = remove_comments_from_element_tree(root)
    validate_xml(root)
    root = transform_into_localnames(root)
    return _replace_namespaces(root, api_url)


def _replace_namespaces(root: etree._Element, api_url: str) -> etree._Element:
    with open("src/dsp_tools/resources/xml_validate/replace_namespace.xslt", "rb") as xslt_file:
        xslt_data = xslt_file.read()
    shortcode = root.attrib["shortcode"]
    default_ontology = root.attrib["default-ontology"]
    namespace = f"{api_url}/ontology/{shortcode}/{default_ontology}/v2#"
    xslt_root = etree.XML(xslt_data)
    transform = etree.XSLT(xslt_root)
    replacement_value = etree.XSLT.strparam(namespace)
    return transform(root, replacementValue=replacement_value).getroot()



from __future__ import annotations

from typing import Any

import pandas as pd
import regex

from dsp_tools.commands.excel2json.new_lists.models.deserialise import ColumnNodes
from dsp_tools.commands.excel2json.new_lists.models.deserialise import Columns
from dsp_tools.models.exceptions import InputError


def get_lang_string_from_column_name(col_str: str, ending: str = r"(\d+|list|comments)") -> str | None:
    """Gets the language tag of one string."""
    if res := regex.search(rf"^(en|de|fr|it|rm)_{ending}$", col_str):
        return res.group(1)
    return None


def get_columns_of_preferred_lang(
    columns: pd.Index[str], preferred_language: str, ending: str = r"(\d+|list|comments)"
) -> list[str]:
    """Get all the columns that contain the preferred language tag."""
    return sorted(col for col in columns if regex.search(rf"^{preferred_language}_{ending}$", col))


def get_column_info(df_cols: pd.Index[Any]) -> Columns:
    """
    Constructs and returns all the columns that should be present in the df based on the languages used.

    Args:
        df_cols: columns of the df

    Returns:
        Object with column info
    """
    hierarchy_nums = get_hierarchy_nums(df_cols)
    all_langs = get_all_languages_for_columns(df_cols)
    preferred_lang = get_preferred_language_from_columns(df_cols)
    node_cols = []
    for n in hierarchy_nums:
        node_cols.append(ColumnNodes(level_num=int(n), columns=[f"{lang}_{n}" for lang in all_langs]))
    comment_cols = []
    if get_all_languages_for_columns(df_cols, r"comments"):
        comment_cols = [f"{lang}_comments" for lang in all_langs]
    list_columns = [f"{lang}_list" for lang in all_langs]
    return Columns(
        preferred_lang=preferred_lang, list_cols=list_columns, comment_cols=comment_cols, node_cols=node_cols
    )


def get_hierarchy_nums(columns: pd.Index[str]) -> list[int]:
    """Get all the numbers that are used in the column names that contain a language tag."""
    return sorted(
        list(set(int(res.group(2)) for x in columns if (res := regex.search(r"^(en|de|fr|it|rm)_(\d+)$", x))))
    )


def get_all_languages_for_columns(columns: pd.Index[str], ending: str = r"(\d+|list|comments)") -> set[str]:
    """Get all the language tags for all the columns."""
    return set(res for x in columns if (res := get_lang_string_from_column_name(x, ending)))


def get_preferred_language_from_columns(columns: pd.Index[str], ending: str = r"(\d+|list|comments)") -> str:
    """Get the language tag of the preferred language."""
    match = {res.group(1) for x in columns if (res := regex.search(rf"^(en|de|fr|it|rm)_{ending}+$", x))}
    if "en" in match:
        return "en"
    elif "de" in match:
        return "de"
    elif "fr" in match:
        return "fr"
    elif "it" in match:
        return "it"
    elif "rm" in match:
        return "rm"
    msg = (
        f"The columns may only contain the languages: 'en', 'de', 'fr', 'it', 'rm'.\n"
        f"The columns are: {" ".join(columns)}"
    )
    raise InputError(msg)



import regex

_iri_pattern = r"^(http)s?://[\w\.\-~]*(:\d{,6})?(/[\w\-~]+)*(#[\w\-~]*)?"
_resource_iri_pattern = r"https?://rdfh.ch/[a-fA-F0-9]{4}/[\w-]{22}"


def is_iri(s: str) -> bool:
    """Checks whether a string is a valid IRI."""
    return regex.fullmatch(_iri_pattern, s) is not None


def is_resource_iri(s: str) -> bool:
    """Checks whether a string is a valid resource IRI."""
    return regex.fullmatch(_resource_iri_pattern, s) is not None



import pytest
from pytest_unordered import unordered

from dsp_tools.commands.xmlupload.models.ontology_lookup_models import _check_if_text_value_property
from dsp_tools.commands.xmlupload.models.ontology_lookup_models import _extract_classes_and_properties_from_onto
from dsp_tools.commands.xmlupload.models.ontology_lookup_models import _get_all_classes_from_onto
from dsp_tools.commands.xmlupload.models.ontology_lookup_models import _get_all_properties_from_onto
from dsp_tools.commands.xmlupload.models.ontology_lookup_models import (
    _get_all_text_value_types_properties_and_from_onto,
)
from dsp_tools.commands.xmlupload.models.ontology_lookup_models import _make_text_value_property_type_lookup
from dsp_tools.commands.xmlupload.models.ontology_lookup_models import _remove_default_prefix
from dsp_tools.commands.xmlupload.models.ontology_lookup_models import _remove_prefixes


class TestGetAllClassesFromJson:
    @staticmethod
    def test_single_class() -> None:
        test_json = [
            {
                "knora-api:isResourceClass": True,
                "rdfs:label": "Segment einer Audio-Ressource",
                "knora-api:canBeInstantiated": True,
                "rdfs:subClassOf": [],
                "@type": "owl:Class",
                "@id": "testonto:AudioSegment",
            },
        ]
        res_cls = _get_all_classes_from_onto(test_json)
        assert res_cls == ["testonto:AudioSegment"]

    @staticmethod
    def test_property() -> None:
        test_json = [
            {
                "rdfs:label": "URI",
                "rdfs:subPropertyOf": {},
                "knora-api:isEditable": True,
                "knora-api:isResourceProperty": True,
                "@type": "owl:ObjectProperty",
                "salsah-gui:guiAttribute": [],
                "knora-api:objectType": {},
                "salsah-gui:guiElement": {},
                "@id": "testonto:hasUri",
            },
        ]
        res_cls = _get_all_classes_from_onto(test_json)
        assert not res_cls

    @staticmethod
    def test_from_json_resources_and_properties() -> None:
        test_json = [
            {
                "knora-api:isResourceClass": True,
                "rdfs:label": "Sequenz einer Audio-Ressource",
                "knora-api:canBeInstantiated": True,
                "rdfs:subClassOf": [],
                "@type": "owl:Class",
                "@id": "testonto:AudioSegment",
            },
            {
                "rdfs:label": "URI",
                "rdfs:subPropertyOf": {},
                "knora-api:isEditable": True,
                "knora-api:isResourceProperty": True,
                "@type": "owl:ObjectProperty",
                "salsah-gui:guiAttribute": [],
                "knora-api:objectType": {},
                "salsah-gui:guiElement": {},
                "@id": "testonto:hasUri",
            },
        ]
        res_cls = _get_all_classes_from_onto(test_json)
        assert res_cls == ["testonto:AudioSegment"]


def test_get_all_properties_from_json_haslinkto() -> None:
    test_json = [
        {
            "rdfs:label": "hasResource",
            "rdfs:subPropertyOf": {},
            "knora-api:isEditable": True,
            "knora-api:isResourceProperty": True,
            "knora-api:isLinkProperty": True,
            "@type": "owl:ObjectProperty",
            "knora-api:objectType": {},
            "salsah-gui:guiElement": {},
            "@id": "testonto:hasResource",
        },
        {
            "knora-api:isLinkValueProperty": True,
            "rdfs:label": "hasResource",
            "rdfs:subPropertyOf": {},
            "knora-api:isEditable": True,
            "knora-api:isResourceProperty": True,
            "@type": "owl:ObjectProperty",
            "knora-api:objectType": {},
            "salsah-gui:guiElement": {},
            "@id": "testonto:hasResourceValue",
        },
    ]
    res_prop = _get_all_properties_from_onto(test_json)
    assert res_prop == ["testonto:hasResource", "testonto:hasResourceValue"]


def test_get_all_properties_from_json_resources_and_properties() -> None:
    test_json = [
        {
            "knora-api:isResourceClass": True,
            "rdfs:label": "Sequenz einer Audio-Ressource",
            "knora-api:canBeInstantiated": True,
            "rdfs:subClassOf": [],
            "@type": "owl:Class",
            "@id": "testonto:AudioSegment",
        },
        {
            "rdfs:label": "URI",
            "rdfs:subPropertyOf": {},
            "knora-api:isEditable": True,
            "knora-api:isResourceProperty": True,
            "@type": "owl:ObjectProperty",
            "salsah-gui:guiAttribute": [],
            "knora-api:objectType": {},
            "salsah-gui:guiElement": {},
            "@id": "testonto:hasUri",
        },
    ]
    res_prop = _get_all_properties_from_onto(test_json)
    assert res_prop == ["testonto:hasUri"]


def test_deserialize_ontology() -> None:
    test_json = [
        {
            "knora-api:isResourceClass": True,
            "rdfs:label": "Annotation",
            "knora-api:canBeInstantiated": True,
            "rdfs:subClassOf": [],
            "rdfs:comment": "A generic class for representing annotations",
            "@type": "owl:Class",
            "@id": "knora-api:Annotation",
        },
        {
            "rdfs:label": "has Link to",
            "rdfs:subPropertyOf": {},
            "knora-api:isEditable": True,
            "knora-api:isResourceProperty": True,
            "@type": "owl:ObjectProperty",
            "knora-api:objectType": {},
            "@id": "knora-api:hasLinkTo",
            "knora-api:subjectType": {},
            "knora-api:isLinkProperty": True,
            "rdfs:comment": "Represents a direct connection between two resources",
        },
    ]
    res_onto = _extract_classes_and_properties_from_onto(test_json)
    assert res_onto.classes == ["Annotation"]
    assert res_onto.properties == ["hasLinkTo"]


def test_remove_prefixes_knora_classes() -> None:
    test_elements = ["knora-api:Annotation", "knora-api:ArchiveFileValue", "knora-api:ArchiveRepresentation"]
    res = _remove_prefixes(test_elements)
    assert unordered(res) == ["Annotation", "ArchiveFileValue", "ArchiveRepresentation"]


def test_remove_prefixes_knora_properties() -> None:
    test_elements = ["knora-api:attachedToUser", "knora-api:deletedBy"]
    res = _remove_prefixes(test_elements)
    assert unordered(res) == ["attachedToUser", "deletedBy"]


def test_get_all_text_value_types_properties_and_from_onto() -> None:
    test_json = [
        {
            "knora-api:isResourceClass": True,
            "rdfs:label": "Annotation",
            "knora-api:canBeInstantiated": True,
            "rdfs:subClassOf": [],
            "rdfs:comment": "A generic class for representing annotations",
            "@type": "owl:Class",
            "@id": "knora-api:Annotation",
        },
        {
            "rdfs:label": "Simple Text",
            "rdfs:subPropertyOf": {},
            "knora-api:isResourceProperty": True,
            "knora-api:objectType": {"@id": "knora-api:TextValue"},
            "salsah-gui:guiElement": {"@id": "salsah-gui:SimpleText"},
            "@id": "onto:hasSimpleText",
        },
        {
            "rdfs:label": "Textarea",
            "rdfs:subPropertyOf": {},
            "knora-api:isResourceProperty": True,
            "knora-api:objectType": {"@id": "knora-api:TextValue"},
            "salsah-gui:guiElement": {"@id": "salsah-gui:Textarea"},
            "@id": "onto:hasTextarea",
        },
        {
            "rdfs:label": "Richtext",
            "rdfs:subPropertyOf": {},
            "knora-api:isResourceProperty": True,
            "knora-api:objectType": {"@id": "knora-api:TextValue"},
            "salsah-gui:guiElement": {"@id": "salsah-gui:Richtext"},
            "@id": "onto:hasRichtext",
        },
        {
            "rdfs:label": "Editor",
            "rdfs:subPropertyOf": {},
            "knora-api:isResourceProperty": True,
            "knora-api:isLinkProperty": True,
            "@type": "owl:ObjectProperty",
            "knora-api:objectType": {"@id": "onto:Person"},
            "salsah-gui:guiElement": {"@id": "salsah-gui:Searchbox"},
            "@id": "onto:hasEditor",
        },
    ]
    res = _get_all_text_value_types_properties_and_from_onto(test_json)
    assert res == [
        ("onto:hasSimpleText", "salsah-gui:SimpleText"),
        ("onto:hasTextarea", "salsah-gui:Textarea"),
        ("onto:hasRichtext", "salsah-gui:Richtext"),
    ]


def test_check_if_text_value_property_cls() -> None:
    test_dict = {
        "knora-api:isResourceClass": True,
        "rdfs:label": "Annotation",
        "knora-api:canBeInstantiated": True,
        "rdfs:subClassOf": [],
        "rdfs:comment": "A generic class for representing annotations",
        "@type": "owl:Class",
        "@id": "knora-api:Annotation",
    }
    res = _check_if_text_value_property(test_dict)
    assert not res


def test_check_if_text_value_property_link_prop() -> None:
    test_dict = {
        "rdfs:label": "Editor",
        "rdfs:subPropertyOf": {},
        "knora-api:isResourceProperty": True,
        "knora-api:isLinkProperty": True,
        "@type": "owl:ObjectProperty",
        "knora-api:objectType": {"@id": "onto:Person"},
        "salsah-gui:guiElement": {"@id": "salsah-gui:Searchbox"},
        "@id": "onto:hasEditor",
    }
    res = _check_if_text_value_property(test_dict)
    assert not res


def test_check_if_text_value_property_text_prop() -> None:
    test_dict = {
        "rdfs:label": "Textarea",
        "rdfs:subPropertyOf": {},
        "knora-api:isResourceProperty": True,
        "knora-api:objectType": {"@id": "knora-api:TextValue"},
        "salsah-gui:guiElement": {"@id": "salsah-gui:Textarea"},
        "@id": "onto:hasTextarea",
    }
    res = _check_if_text_value_property(test_dict)
    assert res


def test_make_text_value_property_gui() -> None:
    test_list = [
        ("onto:hasSimpleText", "salsah-gui:SimpleText"),
        ("onto_other:hasTextarea", "salsah-gui:Textarea"),
        ("onto:hasRichtext", "salsah-gui:Richtext"),
        ("other_onto:hasRichtext", "salsah-gui:Richtext"),
        ("onto:ontoHasSimpleText", "salsah-gui:SimpleText"),
    ]
    res = _make_text_value_property_type_lookup(test_list, "onto")
    assert res.formatted_text_props == {":hasRichtext", "other_onto:hasRichtext", "hasComment", "hasDescription"}
    assert res.unformatted_text_props == {
        ":hasSimpleText",
        "onto_other:hasTextarea",
        ":ontoHasSimpleText",
        "hasTitle",
        "hasKeyword",
    }


class TestRemoveDefaultPrefix:
    def test_not_default(self) -> None:
        res = _remove_default_prefix("onto:property", "default")
        assert res == "onto:property"

    def test_default(self) -> None:
        res = _remove_default_prefix("default:property", "default")
        assert res == ":property"

    def test_not_default_two(self) -> None:
        res = _remove_default_prefix("onto:property_onto", "default")
        assert res == "onto:property_onto"

    def test_default_two(self) -> None:
        res = _remove_default_prefix("default:property_default", "default")
        assert res == ":property_default"


if __name__ == "__main__":
    pytest.main([__file__])



import shutil
import subprocess
from pathlib import Path

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.commands.project.create.project_create import create_project
from dsp_tools.commands.xmlupload.xmlupload import xmlupload
from dsp_tools.models.exceptions import UserError


def _update_possibly_existing_repo(rosetta_folder: Path) -> bool:
    """
    Makes an attempt to pull the latest version of rosetta from GitHub.

    Args:
        rosetta_folder: path to the (possibly existing) clone of the repo

    Returns:
        True if the repo exists and could be updated, False if the repo doesn't exist or couldn't be updated
    """
    is_rosetta_up_to_date = True
    if rosetta_folder.is_dir():
        print(f"Execute 'git pull' in {rosetta_folder}...")
        completed_process = subprocess.run("git pull".split(), cwd=rosetta_folder, check=False)
        if not completed_process or completed_process.returncode != 0:
            print(f"'git pull' failed. Remove '{rosetta_folder}'...")
            shutil.rmtree(rosetta_folder, ignore_errors=True)
            is_rosetta_up_to_date = False
    else:
        is_rosetta_up_to_date = False

    return is_rosetta_up_to_date


def _clone_repo(
    rosetta_folder: Path,
    enclosing_folder: Path,
) -> None:
    """
    Clones the rosetta repo into the enclosing folder.

    Args:
        rosetta_folder: path to the (not yet existing) clone
        enclosing_folder: path to the (existing) destination where rosetta should be cloned into

    Raises:
        UserError: If rosetta cannot be cloned
    """
    print(f"Clone into {rosetta_folder}...")
    cmd = "git clone https://github.com/dasch-swiss/082E-rosetta-scripts.git".split()
    completed_process = subprocess.run(cmd, cwd=enclosing_folder, check=False)
    if not completed_process or completed_process.returncode != 0:
        raise UserError("There was a problem while cloning the rosetta test project")


def _create_json(rosetta_folder: Path) -> bool:
    """
    Creates the rosetta project on the locally running DSP stack.

    Args:
        rosetta_folder: path to the clone

    Returns:
        True if the project could be created without problems, False if something went wrong during the creation process
    """
    print("Execute 'dsp-tools create rosetta.json'...")
    creds = ServerCredentials(server="http://0.0.0.0:3333", user="root@example.com", password="test")
    return create_project(
        project_file_as_path_or_parsed=rosetta_folder / "rosetta.json",
        creds=creds,
        verbose=False,
    )


def _upload_xml(rosetta_folder: Path) -> bool:
    """
    Uplaod the rosetta data on the locally running DSP stack.

    Args:
        rosetta_folder: path to the clone

    Returns:
        True if all data could be uploaded without problems, False if something went wrong during the upload process
    """
    print("Execute 'dsp-tools xmlupload rosetta.xml'...")
    creds = ServerCredentials(
        user="root@example.com",
        password="test",
        server="http://0.0.0.0:3333",
        dsp_ingest_url="http://0.0.0.0:3340",
    )
    return xmlupload(
        input_file=rosetta_folder / "rosetta.xml",
        creds=creds,
        imgdir=str(rosetta_folder),
    )


def upload_rosetta() -> bool:
    """
    This method clones https://github.com/dasch-swiss/082E-rosetta-scripts
    into ~/.dsp-tools/rosetta.
    If the repository is already there, it pulls instead of cloning.
    Then, rosetta.json is created and rosetta.xml uploaded.

    Raises:
        UserError: If the repo cannot be cloned nor pulled

    Returns:
        True if everything went well
    """

    enclosing_folder = Path.home() / Path(".dsp-tools/rosetta")
    enclosing_folder.mkdir(parents=True, exist_ok=True)
    rosetta_folder = enclosing_folder / "082E-rosetta-scripts"

    is_rosetta_up_to_date = _update_possibly_existing_repo(rosetta_folder=rosetta_folder)
    if not is_rosetta_up_to_date:
        _clone_repo(rosetta_folder=rosetta_folder, enclosing_folder=enclosing_folder)

    success1 = _create_json(rosetta_folder=rosetta_folder)
    success2 = _upload_xml(rosetta_folder=rosetta_folder)

    return success1 and success2



from lxml import etree


def test_to_data_rdf(data_xml: etree._Element) -> None:
    res_list = list(data_xml.iterdescendants(tag="resource"))
    all_types = {x.attrib["restype"] for x in res_list}
    assert all_types == {"http://0.0.0.0:3333/ontology/9999/onto/v2#ClassWithEverything"}
    expected_names = {
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testBoolean",
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testColor",
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testSubDate1",
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testDecimalSimpleText",
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testGeoname",
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testIntegerSimpleText",
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testListProp",
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testHasLinkTo",
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testRichtext",
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testTextarea",
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testTimeValue",
        "http://0.0.0.0:3333/ontology/9999/onto/v2#testUriValue",
        "http://api.knora.org/ontology/knora-api/v2#hasColor",
        "http://api.knora.org/ontology/knora-api/v2#isRegionOf",
        "http://api.knora.org/ontology/knora-api/v2#hasGeometry",
        "http://api.knora.org/ontology/knora-api/v2#hasComment",
    }
    assert set(data_xml.xpath("//@name")) == expected_names



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from typing import Optional
from typing import Union
from typing import cast

import regex
from lxml import etree

from dsp_tools.commands.xmlupload.models.formatted_text_value import FormattedTextValue
from dsp_tools.models.exceptions import XmlUploadError


@dataclass(frozen=True)
class XMLProperty:
    """
    Represents a property of a resource in the XML used for data import.

    Attributes:
        name: The name of the property
        valtype: The type of the property
        values: The list of values of the property
    """

    name: str
    valtype: str
    values: list[XMLValue]

    @staticmethod
    def from_node(node: etree._Element, valtype: str, default_ontology: str) -> XMLProperty:
        """
        The factory for the DSP property

        Args:
            node: the property node, p.ex. `<decimal-prop></decimal-prop>`
            valtype: the type of value given by the name of the property node, p.ex. decimal in `<decimal-prop>`
            default_ontology: the name of the ontology

        Raises:
            XmlUploadError: If an upload fails

        Returns:
            The DSP property
        """
        name = XMLProperty._get_name(node, default_ontology)
        if node.tag.endswith("-prop"):
            values = XMLProperty._get_values_from_normal_props(node, valtype)
        else:
            values = [XMLProperty._get_value_from_knora_base_prop(node)]
        return XMLProperty(name, valtype, values)

    @staticmethod
    def _get_name(node: etree._Element, default_ontology: str) -> str:
        # get the property name which is in format namespace:propertyname, p.ex. rosetta:hasName
        orig = node.attrib.get("name")
        if not orig:  # tags like <isSegmentOf> don't have a name attribute
            return f"knora-api:{node.tag}"
        elif ":" not in orig:
            return f"knora-api:{orig}"
        elif orig.startswith(":"):
            # replace an empty namespace with the default ontology name
            return f"{default_ontology}{orig}"
        else:
            return orig

    @staticmethod
    def _get_values_from_normal_props(node: etree._Element, valtype: str) -> list[XMLValue]:
        # parse the subnodes of the property nodes which contain the actual values of the property
        listname = node.attrib.get("list")  # save the list name if given (only for lists)
        values: list[XMLValue] = []
        for subnode in node:
            if subnode.tag == valtype:  # the subnode must correspond to the expected value type
                values.append(XMLValue.from_node(subnode, valtype, listname))
            else:
                raise XmlUploadError(f"ERROR Unexpected tag: '{subnode.tag}'. Property may contain only value tags!")
        return values

    @staticmethod
    def _get_value_from_knora_base_prop(node: etree._Element) -> XMLValue:
        resrefs = set()
        if node.tag.endswith("hasSegmentBounds"):
            value: str | FormattedTextValue = f"{node.attrib["segment_start"]}:{node.attrib["segment_end"]}"
        elif node.tag.endswith(("hasDescription", "hasComment")):
            value = _extract_formatted_text_from_node(node)
            resrefs = value.find_internal_ids()
        else:
            str_orig = "".join(node.itertext())
            value = _cleanup_unformatted_text(str_orig)
        comment = node.attrib.get("comment")
        permissions = node.attrib.get("permissions")
        link_uuid = node.attrib.get("linkUUID")
        return XMLValue(value=value, resrefs=resrefs, comment=comment, permissions=permissions, link_uuid=link_uuid)


@dataclass
class XMLValue:
    """Represents a value of a resource property in the XML used for data import"""

    value: Union[str, FormattedTextValue]
    resrefs: set[str] = field(default_factory=set)
    comment: Optional[str] = None
    permissions: Optional[str] = None
    link_uuid: Optional[str] = None

    @staticmethod
    def from_node(
        node: etree._Element,
        val_type: str,
        listname: Optional[str] = None,
    ) -> XMLValue:
        """Factory method to create an XMLValue from an XML node"""
        value: Union[str, FormattedTextValue] = ""
        resrefs = set()
        comment = node.get("comment")
        permissions = node.get("permissions")
        if val_type == "text" and node.get("encoding") == "xml":
            value = _extract_formatted_text_from_node(node)
            resrefs = value.find_internal_ids()
        elif val_type == "text" and node.get("encoding") == "utf8":
            str_orig = "".join(node.itertext())
            value = _cleanup_unformatted_text(str_orig)
        elif val_type == "list":
            listname = cast(str, listname)
            value = f"{listname}:" + "".join(node.itertext())
        else:
            value = "".join(node.itertext())
        link_uuid = node.attrib.get("linkUUID")  # not all richtexts have a link, so this attribute is optional
        return XMLValue(value=value, resrefs=resrefs, comment=comment, permissions=permissions, link_uuid=link_uuid)


def _extract_formatted_text_from_node(node: etree._Element) -> FormattedTextValue:
    xmlstr = etree.tostring(node, encoding="unicode", method="xml")
    xmlstr = regex.sub(f"<{node.tag}.*?>|</{node.tag}>", "", xmlstr)
    xmlstr = _cleanup_formatted_text(xmlstr)
    return FormattedTextValue(xmlstr)


def _cleanup_formatted_text(xmlstr_orig: str) -> str:
    """
    In a xml-encoded text value from the XML file,
    there may be non-text characters that must be removed.
    This function:
        - replaces (multiple) line breaks by a space
        - replaces multiple spaces or tabstops by a single space (except within `<code>` or `<pre>` tags)

    Args:
        xmlstr_orig: content of the tag from the XML file, in serialized form

    Returns:
        purged string, suitable to be sent to DSP-API
    """
    # replace (multiple) line breaks by a space
    xmlstr = regex.sub("\n+", " ", xmlstr_orig)

    # replace multiple spaces or tabstops by a single space (except within <code> or <pre> tags)
    # the regex selects all spaces/tabstops not followed by </xyz> without <xyz in between.
    # credits: https://stackoverflow.com/a/46937770/14414188
    xmlstr = regex.sub("( {2,}|\t+)(?!(.(?!<(code|pre)))*</(code|pre)>)", " ", xmlstr)

    # remove spaces after <br/> tags (except within <code> tags)
    xmlstr = regex.sub("((?<=<br/?>) )(?!(.(?!<code))*</code>)", "", xmlstr)

    # remove leading and trailing spaces
    xmlstr = xmlstr.strip()

    return xmlstr


def _cleanup_unformatted_text(string_orig: str) -> str:
    """
    In a utf8-encoded text value from the XML file,
    there may be non-text characters that must be removed.
    This function:
        - removes the `<text>` tags
        - replaces multiple spaces or tabstops by a single space

    Args:
        string_orig: original string from the XML file

    Returns:
        purged string, suitable to be sent to DSP-API
    """
    # remove the <text> tags
    string = regex.sub("<text.*?>", "", string_orig)
    string = regex.sub("</text>", "", string)

    # replace multiple spaces or tabstops by a single space
    string = regex.sub(r" {2,}|\t+", " ", string)

    # remove leading and trailing spaces (of every line, but also of the entire string)
    string = "\n".join([s.strip() for s in string.split("\n")])
    return string.strip()


@dataclass(frozen=True)
class XMLBitstream:
    """
    Represents a bitstream object (file) of a resource in the XML used for data import

    Attributes:
        value: The file path of the bitstream object
        permissions: Reference to the set of permissions for the bitstream object
    """

    value: str
    permissions: Optional[str] = None

    @staticmethod
    def from_node(node: etree._Element) -> XMLBitstream:
        """Factory that parses a bitstream node from the XML DOM"""
        if not node.text:
            raise XmlUploadError("Empty bitstream tag")
        return XMLBitstream(node.text, node.get("permissions"))


@dataclass(frozen=True)
class IIIFUriInfo:
    """
    Represents a IIIF URI of a resource in the XML used for data import

    Attributes:
        value: The IIIF URI of the object
        permissions: Reference to the set of permissions for the IIIF URI
    """

    value: str
    permissions: str | None = None

    @staticmethod
    def from_node(node: etree._Element) -> IIIFUriInfo:
        """Factory that parses an IIIF URI node from the XML DOM"""
        if not node.text:
            raise XmlUploadError("Empty IIIF URI tag")
        return IIIFUriInfo(node.text, node.get("permissions"))



from dataclasses import dataclass
from dataclasses import field
from typing import Protocol
from urllib.parse import quote_plus

from dsp_tools.models.exceptions import BaseError
from dsp_tools.models.exceptions import UserError
from dsp_tools.utils.connection import Connection


@dataclass(frozen=True)
class ProjectInfo:
    """Information about a project."""

    project_iri: str
    ontology_iris: list[str]


class ProjectClient(Protocol):
    """Interface (protocol) for project-related requests to the DSP-API."""

    con: Connection
    shortcode: str
    project_info: ProjectInfo | None

    def get_project_iri(self) -> str:
        """Get the IRI of the project to which the data is being uploaded."""

    def get_ontology_name_dict(self) -> dict[str, str]:
        """Returns a mapping of ontology names to ontology IRIs."""


@dataclass()
class ProjectClientLive:
    """Client handling project-related requests to the DSP-API."""

    con: Connection
    shortcode: str
    project_info: ProjectInfo | None = field(init=False, default=None)

    def get_project_iri(self) -> str:
        """Get the IRI of the project to which the data is being uploaded."""
        if not self.project_info:
            self.project_info = _get_project_info_from_server(self.con, self.shortcode)
        return self.project_info.project_iri

    def get_ontology_name_dict(self) -> dict[str, str]:
        """Returns a mapping of ontology names to ontology IRIs."""
        if not self.project_info:
            self.project_info = _get_project_info_from_server(self.con, self.shortcode)
        return {_extract_name_from_onto_iri(iri): iri for iri in self.project_info.ontology_iris}


def _get_project_info_from_server(con: Connection, shortcode: str) -> ProjectInfo:
    project_iri = _get_project_iri_from_server(con, shortcode)
    ontologies = _get_ontologies_from_server(con, project_iri)
    return ProjectInfo(project_iri=project_iri, ontology_iris=ontologies)


def _get_project_iri_from_server(con: Connection, shortcode: str) -> str:
    url = f"/admin/projects/shortcode/{shortcode}"
    try:
        res = con.get(url)
    except BaseError as e:
        raise UserError(f"A project with shortcode {shortcode} could not be found on the DSP server") from e

    iri: str | None = res.get("project", {}).get("id")
    if not iri:
        raise BaseError(f"Unexpected response from server: {res}")
    return iri


def _get_ontologies_from_server(con: Connection, project_iri: str) -> list[str]:
    try:
        iri = quote_plus(project_iri)
        url = f"/v2/ontologies/metadata/{iri}"
        res = con.get(url)
        body = res.get("@graph", res)
        match body:
            case list():
                return [o["@id"] for o in body]
            case dict():
                return [body["@id"]]
            case _:
                raise BaseError("Unexpected response from server")
    except BaseError as e:
        raise BaseError(f"Ontologies for project {project_iri} could not be retrieved from the DSP server") from e


def _extract_name_from_onto_iri(iri: str) -> str:
    return iri.split("/")[-2]



from typing import Callable
from typing import Sequence
from typing import cast

from lxml import etree

from dsp_tools.commands.xml_validate.models.data_deserialised import AbstractResource
from dsp_tools.commands.xml_validate.models.data_deserialised import AnnotationDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import AudioSegmentDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import BooleanValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import ColorValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import DataDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import DateValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import DecimalValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import GeonameValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import IntValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import LinkObjDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import LinkValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import ListValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import ProjectDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import ProjectInformation
from dsp_tools.commands.xml_validate.models.data_deserialised import RegionDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import ResourceDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import RichtextDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import SimpleTextDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import TimeValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import UriValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import ValueDeserialised
from dsp_tools.commands.xml_validate.models.data_deserialised import VideoSegmentDeserialised


def deserialise_xml(root: etree._Element) -> ProjectDeserialised:
    """
    Takes the root of an XML
    Extracts the data of the project and transforms all its resources.

    Args:
        root: root of an xml with qnames and comments removed

    Returns:
        Class instance with the information reformatted
    """
    shortcode = root.attrib["shortcode"]
    default_ontology = root.attrib["default-ontology"]
    project_info = ProjectInformation(shortcode, default_ontology)
    data_deserialised = _deserialise_all_resources(root)
    return ProjectDeserialised(project_info, data_deserialised)


def _deserialise_all_resources(root: etree._Element) -> DataDeserialised:
    all_res: list[AbstractResource] = []
    for res in root.iterchildren():
        res_id = res.attrib["id"]
        lbl = cast(str, res.attrib.get("label"))
        match res.tag:
            case "resource":
                all_res.append(_deserialise_one_resource(res))
            case "annotation":
                all_res.append(AnnotationDeserialised(res_id, lbl))
            case "region":
                all_res.append(RegionDeserialised(res_id, lbl))
            case "link":
                all_res.append(LinkObjDeserialised(res_id, lbl))
            case "video-segment":
                all_res.append(VideoSegmentDeserialised(res_id, lbl))
            case "audio-segment":
                all_res.append(AudioSegmentDeserialised(res_id, lbl))
            case _:
                pass
    return DataDeserialised(all_res)


def _deserialise_one_resource(resource: etree._Element) -> ResourceDeserialised:
    values: list[ValueDeserialised] = []
    for val in resource.iterchildren():
        values.extend(_deserialise_one_property(val))
    return ResourceDeserialised(
        res_id=resource.attrib["id"],
        res_class=resource.attrib["restype"],
        label=resource.attrib["label"],
        values=values,
    )


def _deserialise_one_property(prop_ele: etree._Element) -> Sequence[ValueDeserialised]:  # noqa: PLR0911 (too-many-branches, return statements)
    match prop_ele.tag:
        case "boolean-prop":
            return _deserialise_one_value(prop_ele, BooleanValueDeserialised)
        case "color-prop":
            return _deserialise_one_value(prop_ele, ColorValueDeserialised)
        case "date-prop":
            return _deserialise_one_value(prop_ele, DateValueDeserialised)
        case "decimal-prop":
            return _deserialise_one_value(prop_ele, DecimalValueDeserialised)
        case "geoname-prop":
            return _deserialise_one_value(prop_ele, GeonameValueDeserialised)
        case "list-prop":
            return _deserialise_list_prop(prop_ele)
        case "integer-prop":
            return _deserialise_one_value(prop_ele, IntValueDeserialised)
        case "resptr-prop":
            return _deserialise_one_value(prop_ele, LinkValueDeserialised)
        case "text-prop":
            return _deserialise_text_prop(prop_ele)
        case "time-prop":
            return _deserialise_one_value(prop_ele, TimeValueDeserialised)
        case "uri-prop":
            return _deserialise_one_value(prop_ele, UriValueDeserialised)
        case _:
            return []


def _deserialise_one_value(
    prop: etree._Element, func: Callable[[str, str | None], ValueDeserialised]
) -> Sequence[ValueDeserialised]:
    prop_name = prop.attrib["name"]
    return [func(prop_name, x.text) for x in prop.iterchildren()]


def _deserialise_list_prop(prop: etree._Element) -> list[ListValueDeserialised]:
    prop_name = prop.attrib["name"]
    list_name = prop.attrib["list"]
    return [ListValueDeserialised(prop_name, x.text, list_name) for x in prop.iterchildren()]


def _deserialise_text_prop(prop: etree._Element) -> list[SimpleTextDeserialised | RichtextDeserialised]:
    prop_name = prop.attrib["name"]
    all_vals: list[SimpleTextDeserialised | RichtextDeserialised] = []
    for val in prop.iterchildren():
        match val.attrib["encoding"]:
            case "utf8":
                all_vals.append(SimpleTextDeserialised(prop_name, val.text))
            case "xml":
                all_vals.append(RichtextDeserialised(prop_name, val.text))
    return all_vals



from __future__ import annotations

from dataclasses import dataclass

from rdflib import Graph
from rdflib.term import Node


@dataclass
class ValidationReport:
    conforms: bool
    validation_graph: Graph
    shacl_graph: Graph
    data_graph: Graph


@dataclass
class UnexpectedComponent:
    component_type: str


@dataclass
class ValidationResult:
    source_constraint_component: Node
    res_iri: Node
    res_class: Node
    property: Node
    results_message: str
    value: str | None = None



from __future__ import annotations

import unittest
from typing import Any
from typing import cast

import numpy as np
import pandas as pd
import pytest
import regex
from pandas.testing import assert_frame_equal
from pandas.testing import assert_series_equal
from pytest_unordered import unordered

import dsp_tools.commands.excel2json.utils as utl
from dsp_tools.commands.excel2json.models.input_error import DuplicatesInColumnProblem
from dsp_tools.commands.excel2json.models.ontology import LanguageDict
from dsp_tools.models.exceptions import InputError

# ruff: noqa: PT009 (pytest-unittest-assertion) (remove this line when pytest is used instead of unittest)


def test_find_duplicate_col_names_good() -> None:
    utl._find_duplicate_col_names("file", ["1", "2"])


def test_find_duplicate_col_names_raises() -> None:
    expected = regex.escape(
        "The Excel file 'excelfile' contains the following problems:\n\n"
        "The sheet names inside the same Excel file must be unique. "
        "Using capitalisation or spaces to differentiate sheets is not valid.\n"
        "For example 'sheet' and 'SHEET  ' are considered identical.\n"
        "Under this condition, the following sheet names appear multiple times:\n"
        "    - a\n"
        "    - b"
    )

    with pytest.raises(InputError, match=expected):
        utl._find_duplicate_col_names("excelfile", ["a", "A", "b", "b  ", "c"])


class TestUtils(unittest.TestCase):
    def test_clean_data_frame(self) -> None:
        original_df = pd.DataFrame(
            {
                "  TitLE of Column 1 ": [1.54, " 0-1 ", "1-n ", "   "],
                " Title of Column 2 ": ["1", 1, "   t  ext  ", "text"],
                "Title of Column 3": ["", pd.NA, None, "text"],
            }
        )
        expected_df = pd.DataFrame(
            {
                "title of column 1": ["1.54", "0-1", "1-n", pd.NA],
                "title of column 2": ["1", "1", "t  ext", "text"],
                "title of column 3": [pd.NA, pd.NA, pd.NA, "text"],
            }
        )
        returned_df = utl.clean_data_frame(df=original_df)
        assert_frame_equal(expected_df, returned_df)

    def test_check_contains_required_columns(self) -> None:
        original_df = pd.DataFrame(columns=["col1", "col2", "col3", "extra_col"])
        required = {"col1", "col2", "col3"}
        assert not utl.check_contains_required_columns(df=original_df, required_columns=required)

        required = {"col1", "col2", "col3", "col4"}
        res = utl.check_contains_required_columns(df=original_df, required_columns=required)
        assert res.columns == ["col4"]  # type: ignore[union-attr]

    def test_check_column_for_duplicate(self) -> None:
        original_df = pd.DataFrame(
            {
                "col_1": ["1.54", "0-1", "1-n", "0-1", "1.54"],
                "col_2": ["1.54", "0-1", "1-n", "text", "neu"],
            }
        )
        assert not utl.check_column_for_duplicate(df=original_df, to_check_column="col_2")

        result = utl.check_column_for_duplicate(df=original_df, to_check_column="col_1")
        res = cast(DuplicatesInColumnProblem, result)
        assert res.column == "col_1"
        assert unordered(res.duplicate_values) == ["1.54", "0-1"]

    def test_check_required_values(self) -> None:
        original_df = pd.DataFrame(
            {
                "col_1": ["1.54", "0-1", "1-n", pd.NA],
                "col_2": ["1", "1", pd.NA, "text"],
                "col_3": ["1", "1", "1", "text"],
            }
        )
        returned_dict = utl.check_required_values(df=original_df, required_values_columns=["col_1", "col_3"])
        assert list(returned_dict.keys()) == ["col_1"]
        expected_series = pd.Series([False, False, False, True], name="col_1")
        assert_series_equal(returned_dict["col_1"], expected_series)

    def test_turn_bool_array_into_index_numbers(self) -> None:
        original_series = pd.Series([False, True, False, True])

        returned_list = utl._turn_bool_array_into_index_numbers(series=original_series, true_remains=True)
        assert unordered(returned_list) == [1, 3]

        returned_list = utl._turn_bool_array_into_index_numbers(series=original_series, true_remains=False)
        assert unordered(returned_list) == [0, 2]

    def test_get_wrong_row_numbers(self) -> None:
        original_dict = {
            "col_1": pd.Series([False, True, False, True]),
            "col_2": pd.Series([False, False, True, False]),
        }
        expected_dict = {"col_1": [3, 5], "col_2": [4]}
        returned_dict = utl.get_wrong_row_numbers(wrong_row_dict=original_dict, true_remains=True)
        self.assertDictEqual(expected_dict, returned_dict)

    def test_find_one_full_cell_in_cols(self) -> None:
        required_cols = ["label_en", "label_de", "label_fr", "label_it", "label_rm"]
        original_df = pd.DataFrame(
            {
                "label_en": [1, pd.NA, pd.NA, 4],
                "label_de": [1, pd.NA, 3, 4],
                "label_fr": [1, pd.NA, 3, pd.NA],
                "label_it": [1, pd.NA, 3, 4],
                "label_rm": [pd.NA, pd.NA, 3, 4],
            }
        )
        expected_array = pd.Series([False, True, False, False])
        returned_array = utl.find_one_full_cell_in_cols(df=original_df, required_columns=required_cols)
        assert returned_array is not None
        assert_series_equal(returned_array, expected_array)

    def test_find_one_full_cell_none(self) -> None:
        required_cols = ["label_en", "label_de", "label_fr", "label_it", "label_rm"]
        original_df = pd.DataFrame(
            {
                "label_en": [1, 2, 3, 4],
                "label_de": [1, pd.NA, 3, 4],
                "label_fr": [1, pd.NA, 3, pd.NA],
                "label_it": [1, pd.NA, 3, 4],
                "label_rm": [pd.NA, pd.NA, 3, 4],
            }
        )
        assert not utl.find_one_full_cell_in_cols(df=original_df, required_columns=required_cols)

    def test_col_must_or_not_empty_based_on_other_col(self) -> None:
        original_df = pd.DataFrame({"substring": ["1", "2", "3", "4", "5", "6"], "check": [1, pd.NA, 3, 4, pd.NA, 6]})
        assert not utl.col_must_or_not_empty_based_on_other_col(
            df=original_df,
            substring_list=["1", "3", "6"],
            substring_colname="substring",
            check_empty_colname="check",
            must_have_value=True,
        )

        expected_series = pd.Series([True, False, False, False, False, False])
        returned_series = utl.col_must_or_not_empty_based_on_other_col(
            df=original_df,
            substring_list=["1", "2"],
            substring_colname="substring",
            check_empty_colname="check",
            must_have_value=False,
        )
        assert returned_series is not None
        assert_series_equal(returned_series, expected_series)

    def test_get_labels(self) -> None:
        original_df = pd.DataFrame(
            {
                "label_en": ["text_en", "text_en"],
                "label_de": ["text_de", pd.NA],
                "label_fr": ["text_fr", pd.NA],
                "label_it": ["text_it", pd.NA],
                "label_rm": ["text_rm", pd.NA],
            }
        )
        expected_dict = {"de": "text_de", "en": "text_en", "fr": "text_fr", "it": "text_it", "rm": "text_rm"}
        returned_dict = utl.get_labels(cast("pd.Series[Any]", original_df.loc[0, :])).serialise()
        self.assertDictEqual(expected_dict, returned_dict)

        expected_dict = {"en": "text_en"}
        returned_dict = utl.get_labels(cast("pd.Series[Any]", original_df.loc[1, :])).serialise()
        self.assertDictEqual(expected_dict, returned_dict)

    def test_get_comments(self) -> None:
        original_df = pd.DataFrame(
            {
                "comment_en": ["text_en", pd.NA],
                "comment_de": ["text_de", pd.NA],
                "comment_fr": ["text_fr", pd.NA],
                "comment_it": [pd.NA, pd.NA],
                "comment_rm": ["text_rm", pd.NA],
            }
        )
        expected_dict = {"de": "text_de", "en": "text_en", "fr": "text_fr", "rm": "text_rm"}
        returned_dict = utl.get_comments(cast("pd.Series[Any]", original_df.loc[0, :]))
        assert isinstance(returned_dict, LanguageDict)
        assert expected_dict == returned_dict.serialise()
        assert not utl.get_comments(cast("pd.Series[Any]", original_df.loc[1, :]))


def test_add_optional_columns_with_missing_cols() -> None:
    original_df = pd.DataFrame(
        {
            "comment_en": ["text_en", np.nan],
            "comment_it": ["text_it", np.nan],
            "comment_rm": [np.nan, np.nan],
        }
    )
    expected_df = pd.DataFrame(
        {
            "comment_de": [np.nan, np.nan],
            "comment_en": ["text_en", np.nan],
            "comment_fr": [np.nan, np.nan],
            "comment_it": ["text_it", np.nan],
            "comment_rm": [np.nan, np.nan],
            "label_de": [np.nan, np.nan],
            "label_en": [np.nan, np.nan],
            "label_fr": [np.nan, np.nan],
            "label_it": [np.nan, np.nan],
            "label_rm": [np.nan, np.nan],
            "subject": [np.nan, np.nan],
        }
    )
    optional_col_set = {
        "label_en",
        "label_de",
        "label_fr",
        "label_it",
        "label_rm",
        "comment_en",
        "comment_de",
        "comment_fr",
        "comment_it",
        "comment_rm",
        "subject",
    }
    returned_df = utl.add_optional_columns(original_df, optional_col_set)
    # as the columns are extracted via a set, they are not sorted and may appear in any order,
    # this would cause the validation to fail
    returned_df = returned_df.sort_index(axis=1)
    assert_frame_equal(expected_df, returned_df, check_dtype=False)


def test_add_optional_columns_no_missing_cols() -> None:
    expected_df = pd.DataFrame(
        {
            "comment_de": [pd.NA, pd.NA],
            "comment_en": ["text_en", pd.NA],
            "comment_fr": [pd.NA, pd.NA],
            "comment_it": ["text_it", pd.NA],
            "comment_rm": [pd.NA, pd.NA],
            "label_de": [pd.NA, pd.NA],
            "label_en": [pd.NA, pd.NA],
            "label_fr": [pd.NA, pd.NA],
            "label_it": [pd.NA, pd.NA],
            "label_rm": [pd.NA, pd.NA],
            "subject": [pd.NA, pd.NA],
        }
    )
    unchanged_df = utl.add_optional_columns(expected_df, set())
    assert_frame_equal(expected_df, unchanged_df)


if __name__ == "__main__":
    pytest.main([__file__])



from dataclasses import dataclass


@dataclass(frozen=True)
class ServerCredentials:
    """Contains credentials to connect to a server."""

    user: str
    password: str
    server: str
    dsp_ingest_url: str = "http://0.0.0.0:3340"



from test.unittests.commands.xml_validate.fixtures_data_deserialised import *  # noqa: F403
from test.unittests.commands.xml_validate.fixtures_data_rdf import *  # noqa: F403
from test.unittests.commands.xml_validate.fixtures_onto import *  # noqa: F403
from test.unittests.commands.xml_validate.fixtures_xml_data import *  # noqa: F403



from __future__ import annotations

from copy import deepcopy
from pathlib import Path
from typing import cast

import pandas as pd
from loguru import logger
from lxml import etree

from dsp_tools.commands.ingest_xmlupload.create_resources.user_information import IngestInformation
from dsp_tools.models.exceptions import InputError


def get_mapping_dict_from_file(shortcode: str) -> dict[str, str]:
    """
    This functions returns the information to replace the original filepaths with the identifier from dsp-ingest.

    Args:
        shortcode: Shortcode of the project

    Returns:
        dictionary with original: identifier from dsp-ingest

    Raises:
        InputError: if no file was found
    """
    filepath = Path(f"mapping-{shortcode}.csv")
    if not filepath.is_file():
        raise InputError(f"No mapping CSV file was found at {filepath}.")
    df = pd.read_csv(filepath)
    msg = f"The file '{filepath}' is used to map the internal original filepaths to the internal image IDs."
    print(msg)
    logger.info(msg)
    return dict(zip(df["original"].tolist(), df["derivative"].tolist()))


def replace_filepath_with_internal_filename(
    xml_tree: etree._Element,
    orig_path_2_asset_id: dict[str, str],
) -> tuple[etree._Element, IngestInformation]:
    """
    Replace the original filepaths in the `<bitstream>` tags by the id filenames of the uploaded files.

    Args:
        xml_tree: The parsed original XML tree
        orig_path_2_asset_id: Mapping from original filenames to asset IDs from the mapping.csv

    Returns:
        A copy of the XMl tree, with the replaced filepaths.
        Message informing if all referenced files were uploaded or not.
    """
    no_id_found = []
    used_media_file_paths = []
    new_tree = deepcopy(xml_tree)
    for elem in new_tree.iter():
        if not etree.QName(elem).localname.endswith("bitstream") or not elem.text:
            continue
        img_path_str = elem.text
        if img_path_str not in orig_path_2_asset_id:
            img_path = Path(elem.text)
            img_path_str = str(img_path.relative_to(Path.cwd()) if img_path.is_absolute() else img_path)
            if img_path_str not in orig_path_2_asset_id:
                img_path_str = str(img_path.with_suffix(img_path.suffix.lower()))
            if img_path_str not in orig_path_2_asset_id:
                img_path_str = str(img_path.with_suffix(img_path.suffix.upper()))

        if img_path_str in orig_path_2_asset_id:
            elem.text = orig_path_2_asset_id[img_path_str]
            used_media_file_paths.append(img_path_str)
        else:
            no_id_found.append((cast("etree._Element", elem.getparent()).attrib["id"], str(elem.text)))

    unused_media_paths = [x for x in orig_path_2_asset_id if x not in used_media_file_paths]
    return new_tree, IngestInformation(unused_mediafiles=unused_media_paths, mediafiles_no_id=no_id_found)



import itertools
from dataclasses import dataclass

import pandas as pd

from dsp_tools.commands.xmlupload.models.ontology_lookup_models import TextValueData

separator = "\n    "
list_separator = "\n    - "
medium_separator = "\n\n"
maximum_prints = 50


@dataclass(frozen=True)
class InvalidOntologyElementsInData:
    """This class saves and prints out the information regarding ontology classes and properties
    that are in the XML but not the ontology."""

    classes: list[tuple[str, list[str], str]]
    properties: list[tuple[str, list[str], str]]
    ontos_on_server: list[str]

    def execute_problem_protocol(self) -> tuple[str, pd.DataFrame | None]:
        """
        If there are any elements in properties or classes,
        this method composes an error message.

        Returns:
            the error message and a dataframe with the errors if they exceed 50 or None
        """
        base_msg = (
            f"\nSome property and/or class type(s) used in the XML are unknown.\n"
            f"The ontologies for your project on the server are:{list_separator}"
            f"{list_separator.join(self.ontos_on_server)}{medium_separator}"
        )
        if cls_msg := self._compose_problem_string_for_cls():
            base_msg += cls_msg + "\n"
        if prop_msg := self._compose_problem_string_for_props():
            base_msg += prop_msg + "\n"
        if (
            self._calculate_num_resources(self.classes) + self._calculate_num_resources(self.properties)
            > maximum_prints
        ):
            df = self._get_problems_as_df()
            return base_msg, df
        return base_msg, None

    def _get_problems_as_df(self) -> pd.DataFrame:
        problems = [
            [
                {
                    "problematic type": probs[0],
                    "resource id": x,
                    "problem": probs[2],
                }
                for x in probs[1]
            ]
            for probs in self.classes
        ]
        problems.extend(
            [
                [
                    {
                        "problematic type": probs[0],
                        "resource id": x,
                        "problem": probs[2],
                    }
                    for x in probs[1]
                ]
                for probs in self.properties
            ]
        )
        unpacked: list[dict[str, str]] = list(itertools.chain(*problems))
        return pd.DataFrame.from_records(unpacked)

    def _calculate_num_resources(self, to_count: list[tuple[str, list[str], str]]) -> int:
        return sum((len(x[1]) for x in to_count))

    def _compose_problem_string_for_cls(self) -> str | None:
        if self.classes:
            if self._calculate_num_resources(self.classes) > maximum_prints:
                return "Many resources have an invalid resource type.\nPlease consult the file for details."

            def _format_cls(cls_tup: tuple[str, list[str], str]) -> str:
                ids = list_separator + list_separator.join(cls_tup[1])
                return (
                    f"    Resource Type: '{cls_tup[0]}'{separator}"
                    f"Problem: '{cls_tup[2]}'{separator}"
                    f"Resource ID(s):{ids}"
                )

            problems = [_format_cls(x) for x in self.classes]

            return (
                f"The following resource(s) have an invalid resource type:{medium_separator}"
                + medium_separator.join(problems)
                + "\n"
            )
        else:
            return None

    def _compose_problem_string_for_props(self) -> str | None:
        if self.properties:
            if self._calculate_num_resources(self.properties) > maximum_prints:
                return "Many properties have an invalid resource type.\nPlease consult the file for details."

            def _format_prop(prop_tup: tuple[str, list[str], str]) -> str:
                ids = list_separator + list_separator.join(prop_tup[1])
                return (
                    f"    Property Name: '{prop_tup[0]}'{separator}"
                    f"Problem: '{prop_tup[2]}'{separator}"
                    f"Resource ID(s):{ids}"
                )

            problems = [_format_prop(x) for x in self.properties]
            return (
                f"The following resource(s) have invalid property type(s):{medium_separator}"
                + medium_separator.join(problems)
                + "\n"
            )
        else:
            return None


@dataclass
class InvalidTextValueEncodings:
    """
    This class contains information about resources and the respective properties that have invalid text encodings.

    An invalid encoding would be a property that specifies `knora-api:Richtext` in the ontology,
        but the `<text>` elements use: `<text encoding="utf8">`.
    OR
    A property that specifies `knora-api:Textarea` or `knora-api:SimpleText`
        but the `<text>` elements use: `<text encoding="xml">`.
    """

    problematic_resources: list[TextValueData]

    def execute_problem_protocol(self) -> tuple[str, pd.DataFrame | None]:
        """
        This method composes an error message for the user.

        Returns:
            the error message and a dataframe with the errors if they exceed the maximum allowed print statements
        """
        base_msg = (
            "\nSome text encodings used in the XML data file are not conform with the gui_element "
            "specified in the JSON ontology.\n"
            "Please consult the ontology regarding the assigned gui_elements."
        )
        df = self._get_problems_as_df()
        if len(df) > maximum_prints:
            return base_msg, df
        return base_msg + medium_separator + _make_msg_from_df(df), None

    def _get_problems_as_df(self) -> pd.DataFrame:
        df = pd.DataFrame(
            {
                "Resource ID": [x.resource_id for x in self.problematic_resources],
                "Resource Type": [x.res_type for x in self.problematic_resources],
                "Property Name": [x.property_name for x in self.problematic_resources],
                "Encoding Used": [x.encoding for x in self.problematic_resources],
            }
        )
        return df.sort_values(by=["Resource Type", "Resource ID", "Property Name"], ignore_index=True)


def _make_msg_from_df(df: pd.DataFrame) -> str:
    groups = df.groupby(by="Resource ID")
    return medium_separator.join([_make_msg_for_one_resource(str(_id), res_df) for _id, res_df in groups])


def _make_msg_for_one_resource(res_id: str, res_df: pd.DataFrame) -> str:
    props = res_df["Property Name"].tolist()
    encding = res_df["Encoding Used"].tolist()
    restype = next(iter(res_df["Resource Type"]))
    problems = [f"Property Name: '{p}' -> Encoding Used: '{e}'" for p, e in zip(props, encding)]
    return f"Resource ID: '{res_id}' | Resource Type: '{restype}'{list_separator}{list_separator.join(problems)}"



import pytest

from dsp_tools.commands.xmlupload.iri_resolver import IriResolver


@pytest.fixture
def resolver() -> IriResolver:
    return IriResolver()


def test_empty_resolver(resolver: IriResolver) -> None:
    assert not resolver.non_empty()


def test_non_empty_resolver(resolver: IriResolver) -> None:
    resolver.update("a", "http://example.com/iri#a")
    assert resolver.non_empty()


def test_get_available(resolver: IriResolver) -> None:
    resolver.update("a", "http://example.com/iri#a")
    assert resolver.get("a") == "http://example.com/iri#a"


def test_not_get_unavailable(resolver: IriResolver) -> None:
    resolver.update("a", "http://example.com/iri#a")
    assert resolver.get("b") is None


def test_update_previous_value(resolver: IriResolver) -> None:
    resolver.update("a", "http://example.com/iri#a")
    assert resolver.get("a") == "http://example.com/iri#a"
    resolver.update("a", "http://example.com/iri#aaa")
    assert resolver.get("a") == "http://example.com/iri#aaa"


if __name__ == "__main__":
    pytest.main([__file__])



"""end to end tests for user class"""

import unittest

import pytest

from dsp_tools.commands.project.models.user import User
from dsp_tools.models.langstring import Languages
from dsp_tools.utils.connection import Connection
from dsp_tools.utils.connection_live import ConnectionLive

# ruff: noqa: PT009 (pytest-unittest-assertion) (remove this line when pytest is used instead of unittest)


# define variables for testing
iri_project_0001 = "http://rdfh.ch/projects/0001"
iri_project_0FFF = "http://rdfh.ch/projects/00FF"
iri_group_thing_searcher = "http://rdfh.ch/groups/0001/thing-searcher"
iri_group_images_reviewer = "http://rdfh.ch/groups/00FF/images-reviewer"


class TestUser(unittest.TestCase):
    con: Connection

    def setUp(self) -> None:
        """
        Creates a connection to DSP-API.
        For each test method, a new TestCase instance is created, so setUp() is executed before each test method.
        """
        self.con = ConnectionLive(server="http://0.0.0.0:3333")
        self.con.login(email="root@example.com", password="test")

    def tearDown(self) -> None:
        """
        Logs out from DSP-API.
        For each test method, a new TestCase instance is created, so tearDown() is executed after each test method.
        """
        self.con.logout()

    def test_user_create(self) -> None:
        user = User(
            con=self.con,
            username="wilee1",
            email="wilee.coyote1@canyon.com",
            givenName="Wile E.1",
            familyName="Coyote1",
            password="BeepBeep1",
            lang=Languages.EN,
            status=True,
            sysadmin=True,
            in_projects={iri_project_0001: True},
            in_groups={iri_group_thing_searcher},
        ).create()
        self.assertEqual(user.username, "wilee1")
        self.assertEqual(user.email, "wilee.coyote1@canyon.com")
        self.assertEqual(user.givenName, "Wile E.1")
        self.assertEqual(user.familyName, "Coyote1")
        self.assertTrue(user.status)
        self.assertEqual(user.lang, Languages.EN)
        self.assertTrue(user.sysadmin)
        self.assertEqual(user.in_projects, {iri_project_0001: True})
        self.assertEqual(user.in_groups, {iri_group_thing_searcher})

    def test_user_read_by_iri(self) -> None:
        user = User(con=self.con, iri="http://rdfh.ch/users/normaluser").read()
        self.assertEqual(user.iri, "http://rdfh.ch/users/normaluser")
        self.assertEqual(user.username, "normaluser")
        self.assertEqual(user.familyName, "User")
        self.assertEqual(user.givenName, "Normal")
        self.assertEqual(user.lang, Languages.DE)
        self.assertTrue(user.status)
        self.assertFalse(user.sysadmin)
        self.assertEqual(user.in_projects, {})

    def test_user_create_and_read_by_email(self) -> None:
        User(
            con=self.con,
            username="wilee2",
            email="wilee.coyote2@canyon.com",
            givenName="Wile E.2",
            familyName="Coyote2",
            password="BeepBeep2",
            lang=Languages.EN,
            status=True,
            sysadmin=True,
            in_projects={iri_project_0001: True},
            in_groups={iri_group_thing_searcher},
        ).create()
        user = User(con=self.con, email="wilee.coyote2@canyon.com").read()
        self.assertEqual(user.username, "wilee2")
        self.assertEqual(user.familyName, "Coyote2")
        self.assertEqual(user.givenName, "Wile E.2")
        self.assertEqual(user.lang, Languages.EN)
        self.assertTrue(user.status)
        self.assertTrue(user.sysadmin)
        self.assertEqual(user.in_projects, {iri_project_0001: True})
        self.assertEqual(user.in_groups, {iri_group_thing_searcher})

    def test_user_get_all_users(self) -> None:
        all_users = User.getAllUsers(self.con)
        for user in all_users:
            self.assertIsNotNone(user.iri)


if __name__ == "__main__":
    pytest.main([__file__])



import pytest
from rdflib import RDF
from rdflib import SH
from rdflib import BNode
from rdflib import Graph
from rdflib import Literal
from rdflib import Namespace

from dsp_tools.commands.xml_validate.sparql.resource_shacl import _construct_0_1_cardinality
from dsp_tools.commands.xml_validate.sparql.resource_shacl import _construct_0_n_cardinality
from dsp_tools.commands.xml_validate.sparql.resource_shacl import _construct_1_cardinality
from dsp_tools.commands.xml_validate.sparql.resource_shacl import _construct_1_n_cardinality
from dsp_tools.commands.xml_validate.sparql.resource_shacl import _construct_all_cardinalities
from dsp_tools.commands.xml_validate.sparql.resource_shacl import _construct_resource_nodeshape
from dsp_tools.commands.xml_validate.sparql.resource_shacl import construct_resource_class_node_shape

ONTO = Namespace("http://0.0.0.0:3333/ontology/9999/onto/v2#")
API_SHAPES = Namespace("http://api.knora.org/ontology/knora-api/shapes/v2#")
PREFIXES = """
@prefix knora-api: <http://api.knora.org/ontology/knora-api/v2#> .
@prefix onto: <http://0.0.0.0:3333/ontology/9999/onto/v2#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix salsah-gui: <http://api.knora.org/ontology/salsah-gui/v2#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
"""


@pytest.fixture
def onto_graph() -> Graph:
    g = Graph()
    g.parse("testdata/xml-validate/onto.ttl")
    return g


@pytest.fixture
def one_res_one_prop() -> Graph:
    ttl = f"""{PREFIXES}
    onto:CardOneResource a owl:Class ;
    rdfs:label "Resource with One Cardinality" ;
    knora-api:canBeInstantiated true ;
    knora-api:isResourceClass true ;
    rdfs:subClassOf 
        knora-api:Resource ,
        [ a owl:Restriction ;
            knora-api:isInherited true ;
            owl:maxCardinality 1 ;
            owl:onProperty knora-api:versionDate ],
        [ a owl:Restriction ;
            knora-api:isInherited true ;
            salsah-gui:guiOrder 0 ;
            owl:cardinality 1 ;
            owl:onProperty onto:testBoolean ] .
    
    onto:testBoolean a owl:ObjectProperty ;
        rdfs:label "Test Boolean" ;
        knora-api:isEditable true ;
        knora-api:isResourceProperty true ;
        knora-api:objectType knora-api:BooleanValue ;
        salsah-gui:guiElement salsah-gui:Checkbox ;
        rdfs:subPropertyOf knora-api:hasValue .
    """
    g = Graph()
    g.parse(data=ttl, format="ttl")
    return g


@pytest.fixture
def one_prop() -> Graph:
    ttl = f"""{PREFIXES}
    onto:testBoolean a owl:ObjectProperty ;
        rdfs:label "Test Boolean" ;
        knora-api:isEditable true ;
        knora-api:isResourceProperty true ;
        knora-api:objectType knora-api:BooleanValue ;
        salsah-gui:guiElement salsah-gui:Checkbox ;
        rdfs:subPropertyOf knora-api:hasValue .
    """
    g = Graph()
    g.parse(data=ttl, format="ttl")
    return g


@pytest.fixture
def card_1() -> Graph:
    ttl = f"""{PREFIXES}
    onto:ClassMixedCard a owl:Class ;
        knora-api:isResourceClass true ;
        knora-api:canBeInstantiated true ;
        rdfs:subClassOf [ 
                a owl:Restriction ;
                salsah-gui:guiOrder 0 ;
                owl:cardinality 1 ;
                owl:onProperty onto:testBoolean
                         ] .
    """
    g = Graph()
    g.parse(data=ttl, format="ttl")
    return g


@pytest.fixture
def card_0_1() -> Graph:
    ttl = f"""{PREFIXES}
    onto:ClassMixedCard a owl:Class ;
        knora-api:isResourceClass true ;
        knora-api:canBeInstantiated true ;
        rdfs:subClassOf [ 
                a owl:Restriction ;
                salsah-gui:guiOrder 1 ;
                owl:maxCardinality 1 ;
                owl:onProperty onto:testDecimalSimpleText
                         ] .
    """
    g = Graph()
    g.parse(data=ttl, format="ttl")
    return g


@pytest.fixture
def card_1_n() -> Graph:
    ttl = f"""{PREFIXES}
    onto:ClassMixedCard a owl:Class ;
        knora-api:isResourceClass true ;
        knora-api:canBeInstantiated true ;
        rdfs:subClassOf [ 
                a owl:Restriction ;
                salsah-gui:guiOrder 2 ;
                owl:minCardinality 1 ;
                owl:onProperty onto:testGeoname
                         ] .
    """
    g = Graph()
    g.parse(data=ttl, format="ttl")
    return g


@pytest.fixture
def card_0_n() -> Graph:
    ttl = f"""{PREFIXES}
    onto:ClassMixedCard a owl:Class ;
        knora-api:isResourceClass true ;
        knora-api:canBeInstantiated true ;
        rdfs:subClassOf [ 
                a owl:Restriction ;
                salsah-gui:guiOrder 3 ;
                owl:minCardinality 0 ;
                owl:onProperty onto:testSimpleText
                         ] .
    """
    g = Graph()
    g.parse(data=ttl, format="ttl")
    return g


class TestCheckTripleNumbersOnto:
    def test_nodeshape(self, onto_graph: Graph) -> None:
        result = _construct_resource_nodeshape(onto_graph)
        number_of_resource_classes = 6
        number_of_ignored_properties = 5
        triples_collection_ignored_props = 2 * number_of_ignored_properties * number_of_resource_classes
        triples_cls_nodeshape = 4 * number_of_resource_classes
        assert len(result) == triples_cls_nodeshape + triples_collection_ignored_props

    def test_cardinality_1(self, onto_graph: Graph) -> None:
        result = _construct_1_cardinality(onto_graph)
        number_of_occurrences_in_onto = 2
        triples_card_1 = 7 * number_of_occurrences_in_onto
        assert len(result) == triples_card_1

    def test_cardinality_0_1(self, onto_graph: Graph) -> None:
        result = _construct_0_1_cardinality(onto_graph)
        number_of_occurrences_in_onto = 4
        triples_card_0_1 = 7 * number_of_occurrences_in_onto
        assert len(result) == triples_card_0_1

    def test_cardinality_0_n(self, onto_graph: Graph) -> None:
        result = _construct_0_n_cardinality(onto_graph)
        number_of_occurrences_in_onto = 21  # Inheritance included
        triples_card_0_n = 3 * number_of_occurrences_in_onto
        assert len(result) == triples_card_0_n

    def test_cardinality_1_n(self, onto_graph: Graph) -> None:
        result = _construct_1_n_cardinality(onto_graph)
        number_of_occurrences_in_onto = 1
        triples_card_1_n = 6 * number_of_occurrences_in_onto
        assert len(result) == triples_card_1_n


def test_construct_resource_class_nodeshape(onto_graph: Graph) -> None:
    result = construct_resource_class_node_shape(onto_graph)
    shape_iri = next(result.subjects(SH.targetClass, ONTO.ClassInheritedCardinality))
    assert shape_iri == ONTO.ClassInheritedCardinality_Shape
    shape_iri = next(result.subjects(SH.targetClass, ONTO.ClassMixedCard))
    assert shape_iri == ONTO.ClassMixedCard_Shape
    shape_iri = next(result.subjects(SH.targetClass, ONTO.CardOneResource))
    assert shape_iri == ONTO.CardOneResource_Shape
    shape_iri = next(result.subjects(SH.targetClass, ONTO.ClassWithEverything))
    assert shape_iri == ONTO.ClassWithEverything_Shape
    shape_iri = next(result.subjects(SH.targetClass, ONTO.TestStillImageRepresentation))
    assert shape_iri == ONTO.TestStillImageRepresentation_Shape
    shape_iri = next(result.subjects(SH.targetClass, ONTO.ClassInheritedCardinalityOverwriting))
    assert shape_iri == ONTO.ClassInheritedCardinalityOverwriting_Shape


def test_construct_resource_nodeshape_one_res(one_res_one_prop: Graph) -> None:
    result = _construct_resource_nodeshape(one_res_one_prop)
    subjects = {iri for x in result.triples((None, None, None)) if not isinstance(iri := x[0], BNode)}
    assert len(subjects) == 1
    subject_iri = subjects.pop()
    assert subject_iri == ONTO.CardOneResource_Shape
    node_triples = list(result.triples((subject_iri, None, None)))
    num_triples = 4
    assert len(node_triples) == num_triples
    assert next(result.subjects(RDF.type, SH.NodeShape)) == subject_iri
    assert next(result.subjects(SH.ignoredProperties)) == subject_iri
    assert next(result.objects(subject_iri, SH.closed)) == Literal(True)


def test_construct_resource_nodeshape_no_res(one_prop: Graph) -> None:
    result = _construct_resource_nodeshape(one_prop)
    assert len(result) == 0


class Test1:
    def test_good(self, card_1: Graph) -> None:
        result = _construct_1_cardinality(card_1)
        assert len(result) == 7
        bn = next(result.subjects(RDF.type, SH.PropertyShape))
        shape_iri = next(result.subjects(SH.property, bn))
        assert shape_iri == ONTO.ClassMixedCard_Shape
        assert str(next(result.objects(bn, SH.minCount))) == "1"
        assert str(next(result.objects(bn, SH.maxCount))) == "1"
        assert next(result.objects(bn, SH.path)) == ONTO.testBoolean
        assert next(result.objects(bn, SH.severity)) == SH.Violation
        assert str(next(result.objects(bn, SH.message))) == "1"

    def test_good_link_value(self, link_prop_card_1: Graph) -> None:
        result = _construct_1_cardinality(link_prop_card_1)
        assert len(result) == 7
        bn = next(result.subjects(RDF.type, SH.PropertyShape))
        shape_iri = next(result.subjects(SH.property, bn))
        assert shape_iri == ONTO.ClassMixedCard_Shape
        assert str(next(result.objects(bn, SH.minCount))) == "1"
        assert str(next(result.objects(bn, SH.maxCount))) == "1"
        assert next(result.objects(bn, SH.path)) == ONTO.testHasLinkToCardOneResource
        assert next(result.objects(bn, SH.severity)) == SH.Violation
        assert str(next(result.objects(bn, SH.message))) == "1"

    def test_empty_0_1(self, card_0_1: Graph) -> None:
        result = _construct_1_cardinality(card_0_1)
        assert len(result) == 0

    def test_empty_1_n(self, card_1_n: Graph) -> None:
        result = _construct_1_cardinality(card_1_n)
        assert len(result) == 0

    def test_empty_0_n(self, card_0_n: Graph) -> None:
        result = _construct_1_cardinality(card_0_n)
        assert len(result) == 0


class Test01:
    def test_good(self, card_0_1: Graph) -> None:
        result = _construct_0_1_cardinality(card_0_1)
        assert len(result) == 7
        bn = next(result.subjects(RDF.type, SH.PropertyShape))
        shape_iri = next(result.subjects(SH.property, bn))
        assert shape_iri == ONTO.ClassMixedCard_Shape
        assert str(next(result.objects(bn, SH.minCount))) == "0"
        assert str(next(result.objects(bn, SH.maxCount))) == "1"
        assert next(result.objects(bn, SH.path)) == ONTO.testDecimalSimpleText
        assert next(result.objects(bn, SH.severity)) == SH.Violation
        assert str(next(result.objects(bn, SH.message))) == "0-1"

    def test_good_link_value(self, link_prop_card_01: Graph) -> None:
        result = _construct_0_1_cardinality(link_prop_card_01)
        assert len(result) == 7
        bn = next(result.subjects(RDF.type, SH.PropertyShape))
        shape_iri = next(result.subjects(SH.property, bn))
        assert shape_iri == ONTO.ClassMixedCard_Shape
        assert str(next(result.objects(bn, SH.minCount))) == "0"
        assert str(next(result.objects(bn, SH.maxCount))) == "1"
        assert next(result.objects(bn, SH.path)) == ONTO.testHasLinkToCardOneResource
        assert next(result.objects(bn, SH.severity)) == SH.Violation
        assert str(next(result.objects(bn, SH.message))) == "0-1"

    def test_empty_1(self, card_1: Graph) -> None:
        result = _construct_0_1_cardinality(card_1)
        assert len(result) == 0

    def test_empty_1_n(self, card_1_n: Graph) -> None:
        result = _construct_0_1_cardinality(card_1_n)
        assert len(result) == 0

    def test_empty_0_n(self, card_0_n: Graph) -> None:
        result = _construct_0_1_cardinality(card_0_n)
        assert len(result) == 0


class Test1N:
    def test_good(self, card_1_n: Graph) -> None:
        result = _construct_1_n_cardinality(card_1_n)
        assert len(result) == 6
        bn = next(result.subjects(RDF.type, SH.PropertyShape))
        shape_iri = next(result.subjects(SH.property, bn))
        assert shape_iri == ONTO.ClassMixedCard_Shape
        assert str(next(result.objects(bn, SH.minCount))) == "1"
        assert next(result.objects(bn, SH.path)) == ONTO.testGeoname
        assert next(result.objects(bn, SH.severity)) == SH.Violation
        assert str(next(result.objects(bn, SH.message))) == "1-n"

    def test_empty_1(self, card_1: Graph) -> None:
        result = _construct_1_n_cardinality(card_1)
        assert len(result) == 0

    def test_empty_1_n(self, card_0_1: Graph) -> None:
        result = _construct_1_n_cardinality(card_0_1)
        assert len(result) == 0

    def test_empty_0_n(self, card_0_n: Graph) -> None:
        result = _construct_1_n_cardinality(card_0_n)
        assert len(result) == 0


class Test0N:
    def test_good(self, card_0_n: Graph) -> None:
        result = _construct_0_n_cardinality(card_0_n)
        assert len(result) == 3
        bn = next(result.subjects(RDF.type, SH.PropertyShape))
        shape_iri = next(result.subjects(SH.property, bn))
        assert shape_iri == ONTO.ClassMixedCard_Shape
        assert next(result.objects(bn, SH.path)) == ONTO.testSimpleText

    def test_empty_1_n(self, card_1_n: Graph) -> None:
        result = _construct_0_n_cardinality(card_1_n)
        assert len(result) == 0

    def test_empty_1(self, card_1: Graph) -> None:
        result = _construct_0_n_cardinality(card_1)
        assert len(result) == 0

    def test_empty_0_1(self, card_0_1: Graph) -> None:
        result = _construct_0_n_cardinality(card_0_1)
        assert len(result) == 0


def test_construct_all_cardinalities(one_res_one_prop: Graph) -> None:
    result = _construct_all_cardinalities(one_res_one_prop)
    assert len(result) == 7
    bn = next(result.subjects(RDF.type, SH.PropertyShape))
    shape_iri = next(result.subjects(SH.property, bn))
    assert shape_iri == ONTO.CardOneResource_Shape
    assert str(next(result.objects(bn, SH.minCount))) == "1"
    assert str(next(result.objects(bn, SH.maxCount))) == "1"
    assert next(result.objects(bn, SH.path)) == ONTO.testBoolean
    assert next(result.objects(bn, SH.severity)) == SH.Violation
    assert str(next(result.objects(bn, SH.message))) == "1"


if __name__ == "__main__":
    pytest.main([__file__])



import json
from typing import Any

from pyld import jsonld
from rdflib import Graph
from rdflib import URIRef


def serialise_property_graph(rdf_graph: Graph, prop_name: URIRef) -> dict[str, Any]:
    """
    It serialises an RDF-graph into json-ld,
    which is conform to the format expected by the DSP-API.
    It returns the information about the property and its value(s).
    It is possible that the graph contains different properties.
    The properties do not have to be from the same ontology.

    Args:
        rdf_graph: RDF graph
        prop_name: One property name used in the graph.
                    If there are several it does not matter which.

    Returns:
        A json-ld
    """
    json_graph = _make_json(rdf_graph)
    return _frame_property(json_graph, prop_name)


def _frame_property(serialised_json: list[dict[str, Any]], prop_name: URIRef) -> dict[str, Any]:
    json_frame: dict[str, Any] = {
        str(prop_name): {},
    }
    framed: dict[str, Any] = jsonld.frame(serialised_json, json_frame)
    return framed


def _make_json(rdf_graph: Graph) -> list[dict[str, Any]]:
    graph_bytes = rdf_graph.serialize(format="json-ld", encoding="utf-8")
    json_graph: list[dict[str, Any]] = json.loads(graph_bytes)
    return json_graph



import urllib
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Protocol

import requests
from loguru import logger
from requests import Session
from requests.adapters import HTTPAdapter
from requests.adapters import Retry

from dsp_tools.commands.xmlupload.models.deserialise.deserialise_value import XMLBitstream
from dsp_tools.commands.xmlupload.models.deserialise.xmlresource import BitstreamInfo
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.models.exceptions import BadCredentialsError
from dsp_tools.models.exceptions import PermanentConnectionError
from dsp_tools.utils.logger_config import WARNINGS_SAVEPATH

STATUS_OK = 200
STATUS_UNAUTHORIZED = 401
STATUS_INTERNAL_SERVER_ERROR = 500


@dataclass(frozen=True)
class IngestResponse:
    """Wrapper around `internal_filename`"""

    internal_filename: str


class AssetClient(Protocol):
    """Protocol for asset handling clients."""

    def get_bitstream_info(
        self,
        bitstream: XMLBitstream,
        permissions_lookup: dict[str, Permissions],
        res_label: str,
        res_id: str,
    ) -> tuple[bool, None | BitstreamInfo]:
        """Uploads the file to the ingest server if applicable, and returns the BitstreamInfo.

        Args:
            bitstream: The bitstream to upload.
            permissions_lookup: The permissions lookup.
            res_label: The resource label (for error message in failure case).
            res_id: The resource ID (for error message in failure case).
        """


@dataclass
class DspIngestClientLive(AssetClient):
    """Client for uploading assets to the DSP-Ingest."""

    dsp_ingest_url: str
    token: str
    shortcode: str
    imgdir: str
    session: Session = field(init=False)

    def __post_init__(self) -> None:
        retries = 6
        self.session = Session()
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=0.3,
            allowed_methods=None,  # means all methods
            status_forcelist=[STATUS_INTERNAL_SERVER_ERROR],
        )
        adapter = HTTPAdapter(max_retries=retry)
        self.session.mount("http://", adapter)
        self.session.mount("https://", adapter)

    def _ingest(self, filepath: Path) -> IngestResponse:
        """Uploads a file to the ingest server and returns the IngestResponse.

        This function sends a POST request to the ingest server with the file to upload.
        It will retry the request 6 times in case of a connection, timeout or internal server error.

        After all retry attempts are exhausted it will raise exceptions if the upload failed.
        The http status code is also checked and if it is not 200, a PermanentConnectionError is raised.

        Args:
            filepath: Path to the file to ingest, could be either absolute or relative.

        Raises:
            BadCredentialsError: If the credentials are invalid.
            PermanentConnectionError: If the connection fails.

        Returns:
            IngestResponse: The internal filename of the uploaded file.
        """
        filename = urllib.parse.quote(filepath.name)
        url = f"{self.dsp_ingest_url}/projects/{self.shortcode}/assets/ingest/{filename}"
        headers = {"Authorization": f"Bearer {self.token}", "Content-Type": "application/octet-stream"}
        timeout = 600
        err = f"Failed to ingest {filepath} to '{url}'."
        with open(filepath, "rb") as binary_io:
            try:
                logger.debug(f"REQUEST: POST to {url}, timeout: {timeout}, headers: {headers | {"Authorization": "*"}}")
                res = self.session.post(
                    url=url,
                    headers=headers,
                    data=binary_io,
                    timeout=timeout,
                )
                logger.debug(f"RESPONSE: {res.status_code}: {res.text}")
                if res.status_code == STATUS_OK:
                    return IngestResponse(internal_filename=res.json()["internalFilename"])
                elif res.status_code == STATUS_UNAUTHORIZED:
                    raise BadCredentialsError("Bad credentials")
                else:
                    user_msg = f"{err} See {WARNINGS_SAVEPATH} for more information."
                    print(user_msg)
                    log_msg = f"{err}. Response status code {res.status_code} '{res.text}'"
                    logger.error(log_msg)
                    raise PermanentConnectionError(log_msg)
            except requests.exceptions.RequestException as e:
                raise PermanentConnectionError(f"{err}. {e}") from e

    def get_bitstream_info(
        self,
        bitstream: XMLBitstream,
        permissions_lookup: dict[str, Permissions],
        res_label: str,
        res_id: str,
    ) -> tuple[bool, None | BitstreamInfo]:
        """Uploads a file to the ingest server and returns the BitstreamInfo."""
        try:
            res = self._ingest(Path(self.imgdir) / Path(bitstream.value))
            msg = f"Uploaded file '{bitstream.value}'"
            logger.info(msg)
            permissions = permissions_lookup.get(bitstream.permissions) if bitstream.permissions else None
            return True, BitstreamInfo(bitstream.value, res.internal_filename, permissions)
        except PermanentConnectionError:
            msg = f"Unable to upload file '{bitstream.value}' of resource '{res_label}' ({res_id})"
            logger.opt(exception=True).warning(msg)
            return False, None


@dataclass(frozen=True)
class BulkIngestedAssetClient(AssetClient):
    """Client for handling media info, if the assets were bulk ingested previously."""

    def get_bitstream_info(
        self,
        bitstream: XMLBitstream,
        permissions_lookup: dict[str, Permissions],
        res_label: str,  # noqa: ARG002
        res_id: str,  # noqa: ARG002
    ) -> tuple[bool, BitstreamInfo | None]:
        """Returns the BitstreamInfo of the already ingested file based on the `XMLBitstream.value`."""
        permissions = permissions_lookup.get(bitstream.permissions) if bitstream.permissions else None
        return True, BitstreamInfo(bitstream.value, bitstream.value, permissions)



"""
This module implements the handling (CRUD) of DSP groups.

CREATE:
    * Instantiate a new object of the class Group with all required parameters
    * Call the ``create``-method on the instance

READ:
    * Instantiate a new object with ``iri``(IRI of group) given
    * Call the ``read``-method on the instance
    * Access the information that has been provided to the instance

UPDATE:
    * You need an instance of an existing Project by reading an instance
    * Change the attributes by assigning the new values
    * Call the ``update```method on the instance

DELETE
    * Instantiate a new objects with ``iri``(IRI of group) given, or use any instance that has the iri set
    * Call the ``delete``-method on the instance

"""

from __future__ import annotations

from typing import Any
from typing import Optional
from typing import Union
from urllib.parse import quote_plus

from dsp_tools.commands.project.models.model import Model
from dsp_tools.commands.project.models.project import Project
from dsp_tools.models.exceptions import BaseError
from dsp_tools.models.langstring import LangString
from dsp_tools.utils.connection import Connection


class Group(Model):
    """
    This class represents a DSP group

    Attributes
    ----------

    con : Connection
        A connection instance to a DSP server

    iri : str
        IRI of the group [get only, cannot be modified after creation of instance]

    name : str
        Name of the group

    descriptions : LangString
        Group descriptions in a given language (Languages.EN, Languages.DE, Languages.FR, Languages.IT, Languages.RM).

    project : str | project
        either the IRI of a project [get only, cannot be modified after creation of instance]
        or an valid Project instance

    selfjoin : boolean
        A flag indicating if selfjoin is allowed in this group

    status : boolean
        A flag indicating if the group is active (True) or inactive/makred deleted (False)

    """

    PROJECT_MEMBER_GROUP: str = "http://www.knora.org/ontology/knora-admin#ProjectMember"
    PROJECT_ADMIN_GROUP: str = "http://www.knora.org/ontology/knora-admin#ProjectAdmin"
    PROJECT_SYSTEMADMIN_GROUP: str = "http://www.knora.org/ontology/knora-admin#SystemAdmin"
    ROUTE: str = "/admin/groups"
    ROUTE_SLASH: str = ROUTE + "/"

    _iri: Optional[str]
    _name: str | None
    _descriptions: LangString
    _project: str
    _selfjoin: bool
    _status: bool

    def __init__(
        self,
        con: Connection,
        iri: Optional[str] = None,
        name: Optional[str] = None,
        descriptions: Optional[LangString] = None,
        project: Optional[Union[str, Project]] = None,
        selfjoin: Optional[bool] = None,
        status: Optional[bool] = None,
    ) -> None:
        super().__init__(con)
        self._iri = iri
        self._name = str(name) if name is not None else None
        self._descriptions = LangString(descriptions)
        if project is not None and isinstance(project, Project):
            self._project = project.iri
        else:
            self._project = str(project) if project is not None else None
        self._selfjoin = bool(selfjoin) if selfjoin is not None else None
        self._status = bool(status) if status is not None else None

    @property
    def iri(self) -> Optional[str]:
        return self._iri

    @property
    def name(self) -> Optional[str]:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value
        self._changed.add("name")

    @property
    def descriptions(self) -> LangString:
        return self._descriptions

    @descriptions.setter
    def descriptions(self, value: Optional[LangString]) -> None:
        self._descriptions = LangString(value)
        self._changed.add("descriptions")

    @property
    def project(self) -> str:
        return self._project

    @property
    def selfjoin(self) -> bool:
        return self._selfjoin

    @selfjoin.setter
    def selfjoin(self, value: bool) -> None:
        self._selfjoin = value
        self._changed.add("selfjoin")

    @property
    def status(self) -> bool:
        return self._status

    @status.setter
    def status(self, value: bool) -> None:
        self._status = value
        self._changed.add("status")

    @classmethod
    def fromJsonObj(cls, con: Connection, json_obj: Any) -> Group:
        group_id = json_obj.get("id")
        if group_id is None:
            raise BaseError('Group "id" is missing')
        name = json_obj.get("name")
        if name is None:
            raise BaseError('Group "name" is missing')
        descriptions = LangString.fromJsonObj(json_obj.get("descriptions"))
        tmp = json_obj.get("project")
        if tmp is None:
            raise BaseError('Group "project" is missing')
        project = tmp.get("id")
        if project is None:
            raise BaseError('Group "project" has no "id"')
        selfjoin = json_obj.get("selfjoin")
        if selfjoin is None:
            raise BaseError("selfjoin is missing")
        status = json_obj.get("status")
        if status is None:
            raise BaseError("Status is missing")
        return cls(
            con=con,
            name=name,
            iri=group_id,
            descriptions=descriptions,
            project=project,
            selfjoin=selfjoin,
            status=status,
        )

    def create(self) -> Group:
        jsonobj = self._toJsonObj_create()
        result = self._con.post(Group.ROUTE, jsonobj)
        return Group.fromJsonObj(self._con, result["group"])

    def _toJsonObj_create(self):
        tmp = {}
        if self._name is None:
            raise BaseError("There must be a valid name!")
        tmp["name"] = self._name
        if not self._descriptions.isEmpty():
            tmp["descriptions"] = self._descriptions.toJsonObj()
        if self._project is None:
            raise BaseError("There must be a valid project!")
        tmp["project"] = self._project
        if self._selfjoin is None:
            raise BaseError("There must be a valid value for selfjoin!")
        tmp["selfjoin"] = self._selfjoin
        if self._status is None:
            raise BaseError("There must be a valid value for status!")
        tmp["status"] = self._status
        return tmp

    def read(self) -> Group:
        result = self._con.get(Group.ROUTE_SLASH + quote_plus(self._iri))
        return Group.fromJsonObj(self._con, result["group"])

    def update(self) -> Optional[Group]:
        updated_group = None
        jsonobj = self._toJsonObj_update()
        if jsonobj:
            result = self._con.put(Group.ROUTE_SLASH + quote_plus(self._iri), jsonobj)
            updated_group = Group.fromJsonObj(self._con, result["group"])
        if self._status is not None and "status" in self._changed:
            jsonobj = {"status": self._status}
            result = self._con.put(Group.ROUTE_SLASH + quote_plus(self._iri) + "/status", jsonobj)
            updated_group = Group.fromJsonObj(self._con, result["group"])
        return updated_group

    def _toJsonObj_update(self) -> dict[str, Any]:
        tmp = {}
        if self._name is not None and "name" in self._changed:
            tmp["name"] = self._name
        if not self._descriptions.isEmpty() and "descriptions" in self._changed:
            tmp["descriptions"] = self._descriptions.toJsonObj()
        if self._selfjoin is not None and "selfjoin" in self._changed:
            tmp["selfjoin"] = self._selfjoin
        return tmp

    def delete(self) -> Group:
        result = self._con.delete(Group.ROUTE_SLASH + quote_plus(self._iri))
        return Group.fromJsonObj(self._con, result["group"])

    @staticmethod
    def getAllGroups(con: Connection) -> list[Group]:
        result = con.get(Group.ROUTE)
        return [Group.fromJsonObj(con, group_item) for group_item in result["groups"]]

    @staticmethod
    def getAllGroupsForProject(con: Connection, proj_iri: str) -> list[Group]:
        return [g for g in Group.getAllGroups(con) if g.project == proj_iri]

    def createDefinitionFileObj(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "descriptions": self.descriptions.createDefinitionFileObj(),
            "selfjoin": self.selfjoin,
            "status": self.status,
        }



import pytest
from rdflib import RDF
from rdflib import SH
from rdflib import Graph
from rdflib import Namespace

from dsp_tools.commands.xml_validate.models.input_problems import MaxCardinalityViolation
from dsp_tools.commands.xml_validate.models.input_problems import MinCardinalityViolation
from dsp_tools.commands.xml_validate.models.input_problems import NonExistentCardinalityViolation
from dsp_tools.commands.xml_validate.models.validation import UnexpectedComponent
from dsp_tools.commands.xml_validate.models.validation import ValidationResult
from dsp_tools.commands.xml_validate.reformat_validaton_result import _extract_one_violation
from dsp_tools.commands.xml_validate.reformat_validaton_result import _reformat_one_violation

VALIDATION_PREFIXES = """
    @prefix onto: <http://0.0.0.0:3333/ontology/9999/onto/v2#> .
    @prefix sh: <http://www.w3.org/ns/shacl#> .
    @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
    """

ONTO = Namespace("http://0.0.0.0:3333/ontology/9999/onto/v2#")
DATA = Namespace("http://data/")


@pytest.fixture
def min_count_violation() -> Graph:
    gstr = f"""{VALIDATION_PREFIXES}
    [ a sh:ValidationResult ;
            sh:focusNode <http://data/id_min_card> ;
            sh:resultMessage "1-n" ;
            sh:resultPath onto:testGeoname ;
            sh:resultSeverity sh:Violation ;
            sh:sourceConstraintComponent sh:MinCountConstraintComponent ;
            sh:sourceShape [ ] ] .
    """
    g = Graph()
    g.parse(data=gstr, format="ttl")
    return g


@pytest.fixture
def data() -> Graph:
    gstr = "<http://data/id_min_card> a <http://0.0.0.0:3333/ontology/9999/onto/v2#ClassMixedCard> ."
    g = Graph()
    g.parse(data=gstr, format="ttl")
    return g


@pytest.fixture
def violation_min_card() -> ValidationResult:
    return ValidationResult(
        source_constraint_component=SH.MinCountConstraintComponent,
        res_iri=DATA.id_min_card,
        res_class=ONTO.ClassMixedCard,
        property=ONTO.testGeoname,
        results_message="1-n",
    )


@pytest.fixture
def violation_max_card() -> ValidationResult:
    return ValidationResult(
        source_constraint_component=SH.MaxCountConstraintComponent,
        res_iri=DATA.id_max_card,
        res_class=ONTO.ClassMixedCard,
        property=ONTO.testDecimalSimpleText,
        results_message="0-1",
    )


@pytest.fixture
def violation_closed() -> ValidationResult:
    return ValidationResult(
        source_constraint_component=SH.ClosedConstraintComponent,
        res_iri=DATA.id_closed_constraint,
        res_class=ONTO.CardOneResource,
        property=ONTO.testIntegerSimpleText,
        results_message="Predicate ns3:testIntegerSimpleText is not allowed (closed shape)",
    )


@pytest.fixture
def violation_unknown() -> ValidationResult:
    return ValidationResult(
        source_constraint_component=SH.UniqueLangConstraintComponent,
        res_iri=DATA.id,
        res_class=ONTO.ClassMixedCard,
        property=ONTO.testIntegerSimpleText,
        results_message="This is a constraint that is not checked in the data and should never appear.",
    )


def test_extract_one_violation(min_count_violation: Graph, data: Graph) -> None:
    bn = next(min_count_violation.subjects(RDF.type, SH.ValidationResult))
    result = _extract_one_violation(bn, min_count_violation, data)
    assert result.source_constraint_component == SH.MinCountConstraintComponent
    assert result.res_iri == DATA.id_min_card
    assert result.res_class == ONTO.ClassMixedCard
    assert result.property == ONTO.testGeoname
    assert result.results_message == "1-n"


class TestReformatViolation:
    def test_min(self, violation_min_card: ValidationResult) -> None:
        result = _reformat_one_violation(violation_min_card)
        assert isinstance(result, MinCardinalityViolation)
        assert result.res_id == "id_min_card"
        assert result.res_type == "onto:ClassMixedCard"
        assert result.prop_name == "onto:testGeoname"
        assert result.expected_cardinality == "1-n"

    def test_max(self, violation_max_card: ValidationResult) -> None:
        result = _reformat_one_violation(violation_max_card)
        assert isinstance(result, MaxCardinalityViolation)
        assert result.res_id == "id_max_card"
        assert result.res_type == "onto:ClassMixedCard"
        assert result.prop_name == "onto:testDecimalSimpleText"
        assert result.expected_cardinality == "0-1"

    def test_closed(self, violation_closed: ValidationResult) -> None:
        result = _reformat_one_violation(violation_closed)
        assert isinstance(result, NonExistentCardinalityViolation)
        assert result.res_id == "id_closed_constraint"
        assert result.res_type == "onto:CardOneResource"
        assert result.prop_name == "onto:testIntegerSimpleText"

    def test_unknown(self, violation_unknown: ValidationResult) -> None:
        result = _reformat_one_violation(violation_unknown)
        assert isinstance(result, UnexpectedComponent)
        assert result.component_type == str(SH.UniqueLangConstraintComponent)


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from enum import unique
from typing import Optional


@dataclass(frozen=True)
class OntoIri:
    """
    Holds an ontology IRI

    Attributes:
        iri: the ontology IRI
        hashtag: True if "#" is used to separate elements, False if element name is appended after "/"
    """

    iri: str
    hashtag: bool


ContextType = dict[str, OntoIri]


@unique
class Cardinality(Enum):
    C_1 = "1"
    C_0_1 = "0-1"
    C_1_n = "1-n"
    C_0_n = "0-n"


class WithId:
    """
    Class helper to get json-ld "@id" thingies
    """

    _tmp: str = None

    def __init__(self, obj: Optional[dict[str, str]]):
        if obj is None:
            return
        self._tmp = obj.get("@id")

    def to_string(self) -> Optional[str]:
        return self._tmp
